
__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import threading

import radical.utils as ru

from ... import utils     as rpu
from ... import states    as rps
from ... import constants as rpc


# ------------------------------------------------------------------------------
# 'enum' for RP's spawner types
EXECUTING_NAME_POPEN = "POPEN"
EXECUTING_NAME_SHELL = "SHELL"
EXECUTING_NAME_ABDS  = "ABDS"


# ==============================================================================
#
class AgentExecutingComponent(rpu.Component):
    """
    Manage the creation of CU processes, and watch them until they are completed
    (one way or the other).  The spawner thus moves the unit from
    PendingExecution to Executing, and then to a final state (or PendingStageOut
    of course).
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg):

        rpu.Component.__init__(self, rpc.AGENT_EXECUTING_COMPONENT, cfg)


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Spawner
    #
    @classmethod
    def create(cls, cfg):

        name = cfg['spawner']

        # Make sure that we are the base-class!
        if cls != AgentExecutingComponent:
            raise TypeError("Factory only available to base class!")

        from .popen import Popen
        from .shell import Shell
        from .abds  import ABDS


        try:
            impl = {
                EXECUTING_NAME_POPEN : Popen,
                EXECUTING_NAME_SHELL : Shell,
                EXECUTING_NAME_ABDS  : ABDS
            }[name]

            impl = impl(cfg)
            return impl

        except KeyError:
            raise ValueError("AgentExecutingComponent '%s' unknown or defunct" % name)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os

from .base import LaunchMethod


# ==============================================================================
#
class MPIRunRSH(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, logger):

        LaunchMethod.__init__(self, cfg, logger)

    # --------------------------------------------------------------------------
    #
    def _configure(self):

        # mpirun_rsh (e.g. on Gordon@SDSC, Stampede@TACC)
        if not self._which('mpirun_rsh'):
            raise Exception("mpirun_rsh could not be found")

        # We don't use the full pathname as the user might load a different
        # compiler / MPI library suite from his CU pre_exec that requires
        # the launcher from that version, as experienced on stampede in #572.
        self.launch_command = 'mpirun_rsh'

    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        opaque_slots = cu['opaque_slots']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_cores   = cud['cores']
        task_args    = cud.get('arguments') or []
        task_argstr  = self._create_arg_string(task_args)

        if not 'task_slots' in opaque_slots:
            raise RuntimeError('insufficient information to launch via %s: %s' \
                    % (self.name, opaque_slots))

        task_slots = opaque_slots['task_slots']

        if task_argstr:
            task_command = "%s %s" % (task_exec, task_argstr)
        else:
            task_command = task_exec

        # Extract all the hosts from the slots
        hosts = [slot.split(':')[0] for slot in task_slots]

        # If we have a CU with many cores, we will create a hostfile and pass
        # that as an argument instead of the individual hosts
        if len(hosts) > 42:

            # Create a hostfile from the list of hosts
            hostfile = self._create_hostfile(hosts, impaired=True)
            hosts_string = "-hostfile %s" % hostfile

        else:

            # Construct the hosts_string ('h1 h2 .. hN')
            hosts_string = " ".join(hosts)

        export_vars = ' '.join([var+"=$"+var for var in self.EXPORT_ENV_VARIABLES if var in os.environ])

        mpirun_rsh_command = "%s -np %d %s %s %s" % (
            self.launch_command, task_cores, hosts_string, export_vars, task_command)

        return mpirun_rsh_command, None


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import time
import threading

import radical.utils as ru

from .. import utils     as rpu
from .. import states    as rps
from .. import constants as rpc


# ==============================================================================
#
class Update(rpu.Worker):
    """
    An UpdateWorker pushes CU and Pilot state updates to mongodb.  Its instances
    compete for update requests on the update_queue.  Those requests will be
    triplets of collection name, query dict, and update dict.  Update requests
    will be collected into bulks over some time (BULK_COLLECTION_TIME), to
    reduce number of roundtrips.
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg):

        rpu.Worker.__init__(self, 'AgentUpdateWorker', cfg)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def create(cls, cfg):

        return cls(cfg)


    # --------------------------------------------------------------------------
    #
    def initialize_child(self):

        self._session_id    = self._cfg['session_id']
        self._mongodb_url   = self._cfg['mongodb_url']
        self._pilot_id      = self._cfg['pilot_id']

        _, db, _, _, _      = ru.mongodb_connect(self._mongodb_url)
        self._mongo_db      = db
        self._cinfo         = dict()            # collection cache
        self._lock          = threading.RLock() # protect _cinfo
        self._state_cache   = dict()            # used to preserve state ordering

        self.declare_subscriber('state', 'agent_state_pubsub', self.state_cb)
        self.declare_idle_cb(self.idle_cb, self._cfg.get('bulk_collection_time'))

        # all components use the command channel for control messages
        self.declare_publisher ('command', rpc.AGENT_COMMAND_PUBSUB)

        # communicate successful startup
        self.publish('command', {'cmd' : 'alive',
                                 'arg' : self.cname})


    # --------------------------------------------------------------------------
    #
    def finalize_child(self):

        # communicate finalization
        self.publish('command', {'cmd' : 'final',
                                 'arg' : self.cname})


    # --------------------------------------------------------------------------
    #
    def _ordered_update(self, cu, state, timestamp=None):
        """
        The update worker can receive states for a specific unit in any order.
        If states are pushed straight to theh DB, the state attribute of a unit
        may not reflect the actual state.  This should be avoided by re-ordering
        on the client side DB consumption -- but until that is implemented we
        enforce ordered state pushes to MongoDB.  We do it like this:

          - for each unit arriving in the update worker
            - check if new state is final
              - yes: push update, but never push any update again (only update
                hist)
              - no:
                check if all expected earlier states are pushed already
                - yes: push this state also
                - no:  only update state history
        """

        s2i = {rps.NEW                          :  0,

               rps.PENDING                      :  1,
               rps.PENDING_LAUNCH               :  2,
               rps.LAUNCHING                    :  3,
               rps.PENDING_ACTIVE               :  4,
               rps.ACTIVE                       :  5,

               rps.UNSCHEDULED                  :  6,
               rps.SCHEDULING                   :  7,
               rps.PENDING_INPUT_STAGING        :  8,
               rps.STAGING_INPUT                :  9,
               rps.AGENT_STAGING_INPUT_PENDING  : 10,
               rps.AGENT_STAGING_INPUT          : 11,
               rps.ALLOCATING_PENDING           : 12,
               rps.ALLOCATING                   : 13,
               rps.EXECUTING_PENDING            : 14,
               rps.EXECUTING                    : 15,
               rps.AGENT_STAGING_OUTPUT_PENDING : 16,
               rps.AGENT_STAGING_OUTPUT         : 17,
               rps.PENDING_OUTPUT_STAGING       : 18,
               rps.STAGING_OUTPUT               : 19,

               rps.DONE                         : 20,
               rps.CANCELING                    : 21,
               rps.CANCELED                     : 22,
               rps.FAILED                       : 23
               }
        i2s = {v:k for k,v in s2i.items()}
        s_max = rps.FAILED

        if not timestamp:
            timestamp = rpu.timestamp()

        # we always push state history
        update_dict = {'$push': {
                           'statehistory': {
                               'state'    : state,
                               'timestamp': timestamp}}}
        uid = cu['_id']

      # self._log.debug(" === inp %s: %s" % (uid, state))

        if uid not in self._state_cache:
            self._state_cache[uid] = {'unsent' : list(),
                                      'final'  : False,
                                      'last'   : rps.AGENT_STAGING_INPUT_PENDING} # we get the cu in this state
        cache = self._state_cache[uid]

        # if unit is already final, we don't push state
        if cache['final']:
          # self._log.debug(" === fin %s: %s" % (uid, state))
            return update_dict

        # if unit becomes final, push state and remember it
        if state in [rps.DONE, rps.FAILED, rps.CANCELED]:
            cache['final'] = True
            cache['last']  = state
            update_dict['$set'] = {'state': state}
          # self._log.debug(" === Fin %s: %s" % (uid, state))
            return update_dict

        # check if we have any consecutive list beyond 'last' in unsent
        cache['unsent'].append(state)
      # self._log.debug(" === lst %s: %s %s" % (uid, cache['last'], cache['unsent']))
        state = None
        for i in range(s2i[cache['last']]+1, s2i[s_max]):
          # self._log.debug(" === chk %s: %s in %s" % (uid, i2s[i], cache['unsent']))
            if i2s[i] in cache['unsent']:
                state = i2s[i]
                cache['unsent'].remove(i2s[i])
              # self._log.debug(" === uns %s: %s" % (uid, state))
            else:
              # self._log.debug(" === brk %s: %s" % (uid, state))
                break

        # the max of the consecutive list is set in te update dict...
        if state:
          # self._log.debug(" === set %s: %s" % (uid, state))
            cache['last'] = state
            update_dict['$set'] = {'state': state}

        # record if final state is sent
        if state in [rps.DONE, rps.FAILED, rps.CANCELED]:
          # self._log.debug(" === FIN %s: %s" % (uid, state))
            cache['final'] = True

        return update_dict


    # --------------------------------------------------------------------------
    #
    def _timed_bulk_execute(self, cinfo):

        # is there any bulk to look at?
        if not cinfo['bulk']:
            return False

        now = time.time()
        age = now - cinfo['last']

        # only push if collection time has been exceeded
        if not age > self._cfg['bulk_collection_time']:
            return False

        res = cinfo['bulk'].execute()
        self._log.debug("bulk update result: %s", res)

        self._prof.prof('unit update bulk pushed (%d)' % len(cinfo['uids']), uid=self._pilot_id)
        for entry in cinfo['uids']:
            uid   = entry[0]
            state = entry[1]
            if state:
                self._prof.prof('update', msg='unit update pushed (%s)' % state, uid=uid)
            else:
                self._prof.prof('update', msg='unit update pushed', uid=uid)

        cinfo['last'] = now
        cinfo['bulk'] = None
        cinfo['uids'] = list()

        return True


    # --------------------------------------------------------------------------
    #
    def idle_cb(self):

        action = 0
        with self._lock:
            for cname in self._cinfo:
                action += self._timed_bulk_execute(self._cinfo[cname])

        return bool(action)


    # --------------------------------------------------------------------------
    #
    def state_cb(self, topic, msg):

        cu = msg

        # FIXME: we don't have any error recovery -- any failure to update unit
        #        state in the DB will thus result in an exception here and tear
        #        down the pilot.
        #
        # FIXME: at the moment, the update worker only operates on units.
        #        Should it accept other updates, eg. for pilot states?
        #
        # got a new request.  Add to bulk (create as needed),
        # and push bulk if time is up.
        uid       = cu['_id']
        state     = cu.get('state')
        timestamp = cu.get('state_timestamp', rpu.timestamp())

        self._prof.prof('get', msg="update unit state to %s" % state, uid=uid)

        cbase       = cu.get('cbase',  '.cu')
        query_dict  = cu.get('query')
        update_dict = cu.get('update')

        if not query_dict:
            query_dict  = {'_id' : uid} # make sure unit is not final?
        if not update_dict:
            update_dict = self._ordered_update (cu, state, timestamp)

        # when the unit is about to leave the agent, we also update stdout,
        # stderr exit code etc
        # FIXME: this probably should be a parameter ('FULL') on 'msg'
        if state in [rps.DONE, rps.FAILED, rps.CANCELED, rps.PENDING_OUTPUT_STAGING]:
            if not '$set' in update_dict:
                update_dict['$set'] = dict()
            update_dict['$set']['stdout'   ] = cu.get('stdout')
            update_dict['$set']['stderr'   ] = cu.get('stderr')
            update_dict['$set']['exit_code'] = cu.get('exit_code')

        # check if we handled the collection before.  If not, initialize
        cname = self._session_id + cbase

        with self._lock:
            if not cname in self._cinfo:
                self._cinfo[cname] = {
                        'coll' : self._mongo_db[cname],
                        'bulk' : None,
                        'last' : time.time(),  # time of last push
                        'uids' : list()
                        }


            # check if we have an active bulk for the collection.  If not,
            # create one.
            cinfo = self._cinfo[cname]

            if not cinfo['bulk']:
                cinfo['bulk'] = cinfo['coll'].initialize_ordered_bulk_op()


            # push the update request onto the bulk
            cinfo['uids'].append([uid, state])
            cinfo['bulk'].find  (query_dict) \
                         .update(update_dict)
            self._prof.prof('bulk', msg='bulked (%s)' % state, uid=uid)

            # attempt a timed update
            self._timed_bulk_execute(cinfo)


# ------------------------------------------------------------------------------




#pylint: disable=C0301, C0103, W0212

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"


import datetime

# -----------------------------------------------------------------------------
# The state "struct" encapsulates a state and its timestamp.
class State(object):
    __slots__ = ('state', 'timestamp')

    # ----------------------------------------
    #
    def __init__(self, state, timestamp):
        """Le constructeur.
        """
        self.state = state
        self.timestamp = timestamp

    # ----------------------------------------
    #
    def __eq__(self, other):
        """Comparison operator for 'equal'. Compares 'State' as well 
           as 'str' types.
        """
        if type(other) == str:
            return self.state == other
        else:
            return self.state == other.state

    # ----------------------------------------
    #
    def __ne__(self, other):
        """Comparison operator for 'not equal'. Compares 'State' as well 
           as 'str' types.
        """
        if type(other) == str:
            return not self.__eq__(other)
        else:
            return not self.__eq__(other.state)

    # ----------------------------------------
    #
    def as_dict(self):
        """Returns the state and its timestamp as a Python dictionary.
        """
        return {"state": self.state, "timestamp": self.timestamp}

    # ----------------------------------------
    #
    def __str__(self):
        """Returns the string representation of the state. The string 
           representation is just the state as string without its timestamp.
        """
        return self.state

# -----------------------------------------------------------------------------
# Common States
NEW                          = 'New'
DONE                         = 'Done'
CANCELING                    = 'Canceling'
CANCELED                     = 'Canceled'
FAILED                       = 'Failed'
PENDING                      = 'Pending'

# -----------------------------------------------------------------------------
# ComputePilot States
PENDING_LAUNCH               = 'PendingLaunch'
LAUNCHING                    = 'Launching'
PENDING_ACTIVE               = 'PendingActive'
ACTIVE                       = 'Active'

# -----------------------------------------------------------------------------
# ComputeUnit States
UNSCHEDULED                  = 'Unscheduled'
PENDING_INPUT_STAGING        = 'PendingInputStaging'
STAGING_INPUT                = 'StagingInput'
AGENT_STAGING_INPUT_PENDING  = 'AgentStagingInputPending'
AGENT_STAGING_INPUT          = 'AgentStagingInput'
PENDING_EXECUTION            = 'PendingExecution'
SCHEDULING                   = 'Scheduling'
ALLOCATING_PENDING           = 'AllocatingPending'
ALLOCATING                   = 'Allocating'
EXECUTING_PENDING            = 'ExecutingPending'
EXECUTING                    = 'Executing'
AGENT_STAGING_OUTPUT_PENDING = 'AgentStagingOutputPending'
AGENT_STAGING_OUTPUT         = 'AgentStagingOutput'
PENDING_OUTPUT_STAGING       = 'PendingOutputStaging'
STAGING_OUTPUT               = 'StagingOutput'




#!/usr/bin/env python

"""
.. module:: radical.pilot.agent
   :platform: Unix
   :synopsis: The agent for RADICAL-Pilot.

   The agent gets CUs by means of the MongoDB.
   The execution of CUs by the Agent is (primarily) configured by the
   triplet (LRMS, LAUNCH_METHOD(s), SCHEDULER):
   - The LRMS detects and structures the information about the resources
     available to agent.
   - The Scheduler maps the execution requests of the LaunchMethods to a
     subset of the resources available to the Agent.
     It does not deal with the "presentation" of this subset.
   - The LaunchMethods configure how to execute (regular and MPI) tasks,
     and know about the specific format to specify the subset of resources.


   Structure:
   ----------
   This represents the planned architecture, which is not fully represented in
   code, yet.

     - class Agent
       - represents the whole thing
       - has a set of StageinWorkers  (threads or procs)
       - has a set of StageoutWorkers (threads or procs)
       - has a set of ExecWorkers     (threads or procs)
       - has a set of UpdateWorkers   (threads or procs)
       - has a HeartbeatMonitor       (threads or procs)
       - has a inputstaging  queue
       - has a outputstaging queue
       - has a execution queue
       - has a update queue
       - loops forever
       - in each iteration
         - pulls CU bulks from DB
         - pushes CUs into inputstaging queue or execution queue (based on
           obvious metric)

     class StageinWorker
       - competes for CU input staging requests from inputstaging queue
       - for each received CU
         - performs staging
         - pushes CU into execution queue
         - pushes stage change notification request into update queue

     class StageoutWorker
       - competes for CU output staging requests from outputstaging queue
       - for each received CU
         - performs staging
         - pushes stage change notification request into update queue

     class ExecWorker
       - manages a partition of the allocated cores
         (partition size == max cu size)
       - competes for CU execution reqeusts from execute queue
       - for each CU
         - prepares execution command
         - pushes command to ExecutionEnvironment
         - pushes stage change notification request into update queue

     class Spawner
       - executes CUs according to ExecWorker instruction
       - monitors CU execution (for completion)
       - gets CU execution reqeusts from ExecWorker
       - for each CU
         - executes CU command
         - monitors CU execution
         - on CU completion
           - pushes CU to outputstaging queue (if staging is needed)
           - pushes stage change notification request into update queue

     class UpdateWorker
       - competes for CU state update reqeusts from update queue
       - for each CU
         - pushes state update (collected into bulks if possible)
         - cleans CU workdir if CU is final and cleanup is requested

     Agent
       |
       +--------------------------------------------------------
       |           |              |              |             |
       |           |              |              |             |
       V           V              V              V             V
     ExecWorker* StageinWorker* StageoutWorker* UpdateWorker* HeartbeatMonitor
       |
       +-------------------------------------------------
       |     |               |                |         |
       |     |               |                |         |
       V     V               V                V         V
     LRMS  MPILaunchMethod TaskLaunchMethod Scheduler Spawner


    NOTE:
    -----
      - Units are progressing through the different worker threads, where, in
        general, the unit changes state when transitioning to the next thread.
        The unit ownership thus *defines* the unit state (its owned by the
        InputStagingWorker, it is in StagingInput state, etc), and the state
        update notifications to the DB are merely informational (and can thus be
        asynchron).  The updates need to be ordered though, to reflect valid and
        correct state transition history.


    TODO:
    -----

    - add option to scheduler to ignore core 0 (which hosts the agent process)
    - add LRMS.partition (n) to return a set of partitioned LRMS for partial
      ExecWorkers
    - publish pilot slot history once on shutdown?  Or once in a while when
      idle?  Or push continuously?
    - Schedulers, LRMSs, LaunchMethods, etc need to be made threadsafe, for the
      case where more than one execution worker threads are running.
    - move util functions to rp.utils or r.utils, and pull the from there
    - split the agent into logical components (classes?), and install along with
      RP.
    - add state asserts after `queue.get ()`
    - move mkdir etc from ingest thread to where its used (input staging or
      execution)
    - the structure of the base scheduler should be suitable for both, UMGR
      scheduling and Agent scheduling.  The algs will be different though,
      mostly because the pilots (as targets of the umgr scheduler) have a wait
      queue, but the cores (targets of the agent scheduler) have not.  Is it
      worthwhile to re-use the structure anyway?
    - all stop() method calls need to be replaced with commands which travel
      through the queues.  To deliver commands timely though we either need
      command prioritization (difficult), or need separate command queues...

"""

__copyright__ = "Copyright 2014-2016, http://radical.rutgers.edu"
__license__   = "MIT"

import os
import sys
import copy
import time
import pprint
import signal

import saga                as rs
import radical.utils       as ru
import radical.pilot       as rp
import radical.pilot.utils as rpu



# ------------------------------------------------------------------------------
#
# http://stackoverflow.com/questions/9539052/python-dynamically-changing-base-classes-at-runtime-how-to
#
# Depending on agent architecture (which is specific to the resource type it
# runs on) can switch between different component types: using threaded (when
# running on the same node), multiprocessing (also for running on the same node,
# but avoiding python's threading problems, for the prices of slower queues),
# and remote processes (for running components on different nodes, using zeromq
# queues for communication).
#
# We do some trickery to keep the actual components independent from the actual
# schema:
#
#   - we wrap the different queue types into a rpu.Queue object
#   - we change the base class of the component dynamically to the respective type
#
# This requires components to adhere to the following restrictions:
#
#   - *only* communicate over queues -- no shared data with other components or
#     component instances.  Note that this also holds for example for the
#     scheduler!
#   - no shared data between the component class and it's run() method.  That
#     includes no sharing of queues.
#   - components inherit from base_component, and the constructor needs to
#     register all required component-internal and -external queues with that
#     base class -- the run() method can then transparently retrieve them from
#     there.
#

# this needs git attribute 'ident' set for this file
git_ident = "$Id$"


# ------------------------------------------------------------------------------
#
def pilot_FAILED(mongo_p=None, pilot_uid=None, logger=None, msg=None):

    if logger:
        logger.error(msg)
        logger.error(ru.get_trace())

    print msg
    print ru.get_trace()

    if mongo_p and pilot_uid:

        now = rpu.timestamp()
        out = None
        err = None
        log = None

        try    : out = open('./agent.out', 'r').read()
        except : pass
        try    : err = open('./agent.err', 'r').read()
        except : pass
        try    : log = open('./agent.log', 'r').read()
        except : pass

        msg = [{"message": msg,              "timestamp": now},
               {"message": rpu.get_rusage(), "timestamp": now}]

        mongo_p.update({"_id": pilot_uid},
            {"$pushAll": {"log"         : msg},
             "$push"   : {"statehistory": {"state"     : rp.FAILED,
                                           "timestamp" : now}},
             "$set"    : {"state"       : rp.FAILED,
                          "stdout"      : rpu.tail(out),
                          "stderr"      : rpu.tail(err),
                          "logfile"     : rpu.tail(log),
                          "finished"    : now}
            })

    else:
        if logger:
            logger.error("cannot log error state in database!")

        print "cannot log error state in database!"


# ------------------------------------------------------------------------------
#
def pilot_CANCELED(mongo_p=None, pilot_uid=None, logger=None, msg=None):

    if logger:
        logger.warning(msg)

    print msg

    if mongo_p and pilot_uid:

        now = rpu.timestamp()
        out = None
        err = None
        log = None

        try    : out = open('./agent.out', 'r').read()
        except : pass
        try    : err = open('./agent.err', 'r').read()
        except : pass
        try    : log = open('./agent.log',    'r').read()
        except : pass

        msg = [{"message": msg,              "timestamp": now},
               {"message": rpu.get_rusage(), "timestamp": now}]

        mongo_p.update({"_id": pilot_uid},
            {"$pushAll": {"log"         : msg},
             "$push"   : {"statehistory": {"state"     : rp.CANCELED,
                                           "timestamp" : now}},
             "$set"    : {"state"       : rp.CANCELED,
                          "stdout"      : rpu.tail(out),
                          "stderr"      : rpu.tail(err),
                          "logfile"     : rpu.tail(log),
                          "finished"    : now}
            })

    else:
        if logger:
            logger.error("cannot log cancel state in database!")

        print "cannot log cancel state in database!"


# ------------------------------------------------------------------------------
#
def pilot_DONE(mongo_p=None, pilot_uid=None, logger=None, msg=None):

    if mongo_p and pilot_uid:

        now = rpu.timestamp()
        out = None
        err = None
        log = None

        try    : out = open('./agent.out', 'r').read()
        except : pass
        try    : err = open('./agent.err', 'r').read()
        except : pass
        try    : log = open('./agent.log',    'r').read()
        except : pass

        msg = [{"message": "pilot done",     "timestamp": now},
               {"message": rpu.get_rusage(), "timestamp": now}]

        mongo_p.update({"_id": pilot_uid},
            {"$pushAll": {"log"         : msg},
             "$push"   : {"statehistory": {"state"    : rp.DONE,
                                           "timestamp": now}},
             "$set"    : {"state"       : rp.DONE,
                          "stdout"      : rpu.tail(out),
                          "stderr"      : rpu.tail(err),
                          "logfile"     : rpu.tail(log),
                          "finished"    : now}
            })

    else:
        if logger:
            logger.error("cannot log cancel state in database!")

        print "cannot log cancel state in database!"


# ==============================================================================
#
# Agent bootstrap stage 3
#
# ==============================================================================
#
def start_bridges(cfg, log):
    """
    For all bridges defined on this agent instance, create that bridge.
    Keep a handle around for shutting them down later.
    """

    log.debug('start_bridges')

    # ----------------------------------------------------------------------
    # shortcut for bridge creation
    bridge_type = {rp.AGENT_STAGING_INPUT_QUEUE  : 'queue',
                   rp.AGENT_SCHEDULING_QUEUE     : 'queue',
                   rp.AGENT_EXECUTING_QUEUE      : 'queue',
                   rp.AGENT_STAGING_OUTPUT_QUEUE : 'queue',
                   rp.AGENT_UNSCHEDULE_PUBSUB    : 'pubsub',
                   rp.AGENT_RESCHEDULE_PUBSUB    : 'pubsub',
                   rp.AGENT_COMMAND_PUBSUB       : 'pubsub',
                   rp.AGENT_STATE_PUBSUB         : 'pubsub'}

    def _create_bridge(name):
        if bridge_type[name] == 'queue':
            return rpu.Queue.create(rpu.QUEUE_ZMQ, name, rpu.QUEUE_BRIDGE)
        elif bridge_type[name] == 'pubsub':
            return rpu.Pubsub.create(rpu.PUBSUB_ZMQ, name, rpu.PUBSUB_BRIDGE)
        else:
            raise ValueError('unknown bridge type for %s' % name)
    # ----------------------------------------------------------------------

    # create all bridges we need.  Use the default addresses,
    # ie. they will bind to all local interfacces on ports 10.000++.
    bridges = dict()
    sub_cfg = cfg['agent_layout']['agent_0']
    for b in sub_cfg.get('bridges', []):

        bridge     = _create_bridge(b)
        bridge_in  = bridge.bridge_in
        bridge_out = bridge.bridge_out
        bridges[b] = {'handle' : bridge,
                      'in'     : bridge_in,
                      'out'    : bridge_out,
                      'alive'  : True}  # no alive check done, yet
        log.info('created bridge %s: %s', b, bridge.name)

    log.debug('start_bridges done')

    return bridges


# --------------------------------------------------------------------------
#
def write_sub_configs(cfg, bridges, nodeip, log):
    """
    create a sub_config for each sub-agent we intent to spawn
    """

    # get bridge addresses from our bridges, and append them to the config
    if not 'bridge_addresses' in cfg:
        cfg['bridge_addresses'] = dict()

    for b in bridges:
        # to avoid confusion with component input and output, we call bridge
        # input a 'sink', and a bridge output a 'source' (from the component
        # perspective)
        sink   = ru.Url(bridges[b]['in'])
        source = ru.Url(bridges[b]['out'])

        # we replace the ip address with what we got from LRMS (nodeip).  The
        # bridge should be listening on all interfaces, but we want to make sure
        # the sub-agents connect on an IP which is accessible to them
        sink.host   = nodeip
        source.host = nodeip

        # keep the resultin URLs as strings, to be used as addresses
        cfg['bridge_addresses'][b] = dict()
        cfg['bridge_addresses'][b]['sink']   = str(sink)
        cfg['bridge_addresses'][b]['source'] = str(source)

    # write deep-copies of the config (with the corrected agent_name) for each
    # sub-agent (apart from agent_0, obviously)
    for sa in cfg.get('agent_layout'):
        if sa != 'agent_0':
            sa_cfg = copy.deepcopy(cfg)
            sa_cfg['agent_name'] = sa
            ru.write_json(sa_cfg, './%s.cfg' % sa)


# --------------------------------------------------------------------------
#
# avoid undefined vars on finalization / signal handling
bridges = dict()
agent   = None
lrms    = None

def bootstrap_3():
    """
    This method continues where the bootstrapper left off, but will quickly pass
    control to the Agent class which will spawn the functional components.

    Most of bootstrap_3 applies only to agent_0, in particular all mongodb
    interactions remains excluded for other sub-agent instances.

    The agent interprets a config file, which will specify in an agent_layout
    section:
      - what nodes should be used for sub-agent startup
      - what bridges should be started
      - what components should be started
      - what are the endpoints for bridges which are not started
    bootstrap_3 will create derived config files for all sub-agents.

    The agent master (agent_0) will collect information about the nodes required
    for all instances.  That is added to the config itself, for the benefit of
    the LRMS initialisation which is expected to block those nodes from the
    scheduler.
    """

    global lrms, agent, bridges

    # find out what agent instance name we have
    if len(sys.argv) != 2:
        raise RuntimeError('invalid number of parameters (%s)' % sys.argv)
    agent_name = sys.argv[1]

    # load the agent config, and overload the config dicts
    agent_cfg  = "%s/%s.cfg" % (os.getcwd(), agent_name)
    print "startup agent %s : %s" % (agent_name, agent_cfg)

    cfg = ru.read_json_str(agent_cfg)
    cfg['agent_name'] = agent_name
    pilot_id = cfg['pilot_id']

    # set up a logger and profiler
    prof = rpu.Profiler ('%s.bootstrap_3' % agent_name)
    prof.prof('sync ref', msg='agent start', uid=pilot_id)
    log  = ru.get_logger('%s.bootstrap_3' % agent_name,
                         '%s.bootstrap_3.log' % agent_name, 'DEBUG')  # FIXME?
    log.info('start')
    prof.prof('sync ref', msg='agent start')

    try:
        import setproctitle as spt
        spt.setproctitle('radical.pilot %s' % agent_name)
    except Exception as e:
        log.debug('no setproctitle: %s', e)

    log.setLevel(cfg.get('debug', 'INFO'))

    print "Agent config (%s):\n%s\n\n" % (agent_cfg, pprint.pformat(cfg))

    # quickly set up a mongodb handle so that we can report errors.
    # FIXME: signal handlers need mongo_p, but we won't have that until later
    if agent_name == 'agent_0':

        # Check for the RADICAL_PILOT_DB_HOSTPORT env var, which will hold the
        # address of the tunnelized DB endpoint.
        # If it exists, we overrule the agent config with it.
        hostport = os.environ.get('RADICAL_PILOT_DB_HOSTPORT')
        if hostport:
            dburl = ru.Url(cfg['mongodb_url'])
            dburl.host, dburl.port = hostport.split(':')
            cfg['mongodb_url'] = str(dburl)

        _, mongo_db, _, _, _  = ru.mongodb_connect(cfg['mongodb_url'])
        mongo_p = mongo_db["%s.p" % cfg['session_id']]

        if not mongo_p:
            raise RuntimeError('could not get a mongodb handle')


    # set up signal and exit handlers
    def exit_handler():
        global lrms, agent, bridges

        print 'atexit'
        if lrms:
            lrms.stop()
            lrms = None
        if bridges:
            for b in bridges:
                b.stop()
            bridges = dict()
        if agent:
            agent.stop()
            agent = None
        sys.exit(1)

    def sigint_handler(signum, frame):
        if agent_name == 'agent_0':
            pilot_FAILED(msg='Caught SIGINT. EXITING (%s)' % frame)
        print 'sigint'
        prof.prof('stop', msg='sigint_handler', uid=pilot_id)
        prof.close()
        sys.exit(2)

    def sigterm_handler(signum, frame):
        if agent_name == 'agent_0':
            pilot_FAILED(msg='Caught SIGTERM. EXITING (%s)' % frame)
        print 'sigterm'
        prof.prof('stop', msg='sigterm_handler %s' % os.getpid(), uid=pilot_id)
        prof.close()
        sys.exit(3)

    def sigalarm_handler(signum, frame):
        if agent_name == 'agent_0':
            pilot_FAILED(msg='Caught SIGALRM (Walltime limit?). EXITING (%s)' % frame)
        print 'sigalrm'
        prof.prof('stop', msg='sigalarm_handler', uid=pilot_id)
        prof.close()
        sys.exit(4)

    import atexit
    atexit.register(exit_handler)
    signal.signal(signal.SIGINT,  sigint_handler)
    signal.signal(signal.SIGTERM, sigterm_handler)
    signal.signal(signal.SIGALRM, sigalarm_handler)

    # if anything went wrong up to this point, we would have been unable to
    # report errors into mongodb.  From here on, any fatal error should result
    # in one of the above handlers or exit handlers being activated, thus
    # reporting the error dutifully.

    try:
        # ----------------------------------------------------------------------
        # des Pudels Kern: merge LRMS info into cfg and get the agent started

        if agent_name == 'agent_0':

            # only the master agent creates LRMS and sub-agent config files.
            # The LRMS which will give us the set of agent_nodes to use for
            # sub-agent startup.  Add the remaining LRMS information to the
            # config, for the benefit of the scheduler).

            lrms = rp.agent.RM.create(name   = cfg['lrms'],
                             cfg    = cfg,
                             logger = log)
            cfg['lrms_info'] = lrms.lrms_info


            # the master agent also is the only one which starts bridges.  It
            # has to do so before creating the Agent Worker instance, as that is
            # using the bridges already.

            bridges = start_bridges(cfg, log)
            # FIXME: make sure all communication channels are in place.  This could
            # be replaced with a proper barrier, but not sure if that is worth it...
            time.sleep (1)

            # after we started bridges, we'll add their in and out addresses
            # to the config, so that the communication channels can connect to
            # them.  At this point we also write configs for all sub-agents this
            # instance intents to spawn.
            #
            # FIXME: we should point the address to the node of the subagent
            #        which hosts the bridge, not the local IP.  Until this
            #        is fixed, bridges MUST run on agent_0 (which is what
            #        RM.hostip() below will point to).
            nodeip = rp.agent.RM.hostip(cfg.get('network_interface'), logger=log)
            write_sub_configs(cfg, bridges, nodeip, log)

            # Store some runtime information into the session
            if 'version_info' in lrms.lm_info:
                mongo_p.update({"_id": pilot_id},
                               {"$set": {"lm_info": lrms.lm_info['version_info']}})

        # we now have correct bridge addresses added to the agent_0.cfg, and all
        # other agents will have picked that up from their config files -- we
        # can start the agent and all its components!
        agent = rp.worker.Agent(cfg)
        agent.start()

        log.debug('waiting for agent %s to join' % agent_name)
        agent.join()
        log.debug('agent %s joined' % agent_name)

        # ----------------------------------------------------------------------

    except SystemExit:
        log.exception("Exit running agent: %s" % agent_name)
        if agent and not agent.final_cause:
            agent.final_cause = "sys.exit"

    except Exception as e:
        log.exception("Error running agent: %s" % agent_name)
        if agent and not agent.final_cause:
            agent.final_cause = "error"

    finally:

        # in all cases, make sure we perform an orderly shutdown.  I hope python
        # does not mind doing all those things in a finally clause of
        # (essentially) main...
        if agent:
            agent.stop()
        log.debug('agent %s finalized' % agent_name)

        # agent.stop will not tear down bridges -- we do that here at last
        for name,b in bridges.items():
            try:
                log.info("closing bridge %s", b)
                b['handle'].stop()
            except Exception as e:
                log.exception('ignore failing bridge terminate (%s)', e)
        bridges = dict()

        # make sure the lrms release whatever it acquired
        if lrms:
            lrms.stop()
            lrms = None

        # agent_0 will also report final pilot state to the DB
        if agent_name == 'agent_0':
            if agent and agent.final_cause == 'timeout':
                pilot_DONE(mongo_p, pilot_id, log, "TIMEOUT received. Terminating.")
            elif agent and agent.final_cause == 'cancel':
                pilot_CANCELED(mongo_p, pilot_id, log, "CANCEL received. Terminating.")
            elif agent and agent.final_cause == 'sys.exit':
                pilot_CANCELED(mongo_p, pilot_id, log, "EXIT received. Terminating.")
            elif agent and agent.final_cause == 'finalize':
                log.info('shutdown due to component finalization -- assuming error')
                pilot_FAILED(mongo_p, pilot_id, log, "FINALIZE received")
            elif agent:
                pilot_FAILED(mongo_p, pilot_id, log, "TERMINATE received")
            else:
                pilot_FAILED(mongo_p, pilot_id, log, "FAILED startup")

        log.info('stop')
        prof.prof('stop', msg='finally clause agent', uid=pilot_id)
        prof.close()


# ==============================================================================
#
if __name__ == "__main__":

    print "---------------------------------------------------------------------"
    print
    print "PYTHONPATH: %s"  % sys.path
    print "python: %s"      % sys.version
    print "utils : %-5s : %s" % (ru.version_detail, ru.__file__)
    print "saga  : %-5s : %s" % (rs.version_detail, rs.__file__)
    print "pilot : %-5s : %s" % (rp.version_detail, rp.__file__)
    print "        type  : multicore"
    print "        gitid : %s" % git_ident
    print
    print "---------------------------------------------------------------------"
    print

    bootstrap_3()

    print "bootstrap_3 done"

#
# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import sys
import radical.pilot as rp
import radical.utils as ru

dh = ru.DebugHelper ()

CNT       = 0

FLOPS     = 10 * 1000 * 1000 * 1000
BYTES_IN  = 10 * 1000 * 1000  # not yet supported
BYTES_OUT = 10 * 1000 * 1000  # not yet supported
BYTES_MEM = 10 * 1000 * 1000  # not yet supported

RUNTIME   = 10
UNITS     = 10
CORES     = 12

SCHED     = rp.SCHED_DIRECT_SUBMISSION

RESOURCE  = 'local.localhost'
PROJECT   = None
QUEUE     = None
SCHEMA    = None
  
# RESOURCE  = 'home.test'
# PROJECT   = None
# QUEUE     = None
# SCHEMA    = 'ssh'

# RESOURCE  = 'epsrc.archer'
# PROJECT   = 'e290'
# QUEUE     = 'short'
# SCHEMA    = None

# RESOURCE  = 'lrz.supermuc'
# PROJECT   = 'e290'
# QUEUE     = 'short'
# SCHEMA    = None

# RESOURCE  = 'xsede.stampede'
# PROJECT   = 'TG-MCB090174' 
# QUEUE     = 'development'
# SCHEMA    = None

# RESOURCE  = 'xsede.gordon'
# PROJECT   = None
# QUEUE     = 'debug'
# SCHEMA    = None

# RESOURCE  = 'xsede.blacklight'
# PROJECT   = None
# QUEUE     = 'debug'
# SCHEMA    = 'gsissh'

# RESOURCE  = 'xsede.trestles'
# PROJECT   = 'TG-MCB090174' 
# QUEUE     = 'shared'
# SCHEMA    = None

# RESOURCE  = 'futuregrid.india'
# PROJECT   = None
# QUEUE     = None
# SCHEMA    = None
  
# RESOURCE  = 'nersc.hopper'
# PROJECT   = None
# QUEUE     = 'debug'
# SCHEMA    = 'ssh'

#------------------------------------------------------------------------------
#
def pilot_state_cb (pilot, state):

    if not pilot:
        return

    print "[Callback]: ComputePilot '%s' state: %s." % (pilot.uid, state)

    if state == rp.FAILED:
        sys.exit (1)


#------------------------------------------------------------------------------
#
def unit_state_cb (unit, state):

    if not unit:
        return

    global CNT

    print "[Callback]: unit %s on %s: %s." % (unit.uid, unit.pilot_id, state)

    if state in [rp.FAILED, rp.DONE, rp.CANCELED]:
        CNT += 1
        print "[Callback]: # %6d" % CNT


    if state == rp.FAILED:
        print "stderr: %s" % unit.stderr
        sys.exit(2)


#------------------------------------------------------------------------------
#
def wait_queue_size_cb(umgr, wait_queue_size):

    print "[Callback]: wait_queue_size: %s." % wait_queue_size


#------------------------------------------------------------------------------
#
if __name__ == "__main__":

    # we can optionally pass session name to RP
    if len(sys.argv) > 1:
        session_name = sys.argv[1]
    else:
        session_name = None

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session(name=session_name)
    print "session id: %s" % session.uid

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        pmgr = rp.PilotManager(session=session)
        pmgr.register_callback(pilot_state_cb)

        pdesc = rp.ComputePilotDescription()
        pdesc.resource      = RESOURCE
        pdesc.cores         = CORES
        pdesc.project       = PROJECT
        pdesc.queue         = QUEUE
        pdesc.runtime       = RUNTIME
        pdesc.cleanup       = False
        pdesc.access_schema = SCHEMA

        pilot = pmgr.submit_pilots(pdesc)

        input_sd_pilot = {
                'source': 'file:///etc/passwd',
                'target': 'staging:///f1',
                'action': rp.TRANSFER
                }
        pilot.stage_in (input_sd_pilot)

        umgr = rp.UnitManager(session=session, scheduler=SCHED)
        umgr.register_callback(unit_state_cb,      rp.UNIT_STATE)
        umgr.register_callback(wait_queue_size_cb, rp.WAIT_QUEUE_SIZE)
        umgr.add_pilots(pilot)

        input_sd_umgr   = {'source':'/etc/group',        'target': 'f2',                'action': rp.TRANSFER}
        input_sd_agent  = {'source':'staging:///f1',     'target': 'f1',                'action': rp.COPY}
        output_sd_agent = {'source':'f1',                'target': 'staging:///f1.bak', 'action': rp.COPY}
        output_sd_umgr  = {'source':'f2',                'target': 'f2.bak',            'action': rp.TRANSFER}


        # we create one pseudo unit which installs radical.synapse in the pilot
        # ve
        cud = rp.ComputeUnitDescription()
        cud.pre_exec    = ["pip uninstall -y radical.synapse",
                           "pip install --upgrade radical.synapse"]
        cud.executable  = "radical-synapse-version"
        cud.cores       = 1

        cu = umgr.submit_units(cud)
        umgr.wait_units(cu.uid)
        assert(cu.state == rp.DONE)

        cuds = list()
        for n in range(1,UNITS+1):
            cud = rp.ComputeUnitDescription()
            cud.executable     = "radical-synapse-sample"
            cud.arguments      = ("-m sample -f %s -s %d" % (FLOPS, n)).split()
            cud.cores          = n
            cud.input_staging  = [ input_sd_umgr,  input_sd_agent]
            cud.output_staging = [output_sd_umgr, output_sd_agent]
            cuds.append(cud)

        units = umgr.submit_units(cuds)

        umgr.wait_units()

        for cu in units:
            print "* Task %s state %s, exit code: %s, started: %s, finished: %s" \
                % (cu.uid, cu.state, cu.exit_code, cu.start_time, cu.stop_time)
            print "out:"
            print cu.stdout
            print "err:"
            print cu.stderr
            print

      # os.system ("radicalpilot-stats -m stat,plot -s %s > %s.stat" % (session.uid, session_name))


    except Exception as e:
        # Something unexpected happened in the pilot code above
        print "caught Exception: %s" % e
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        print "need to exit now: %s" % e

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        print "closing session"
        session.close ()

        # the above is equivalent to
        #
        #   session.close (cleanup=True, terminate=True)
        #
        # it will thus both clean out the session's database record, and kill
        # all remaining pilots (none in our example).


#-------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


from .base import LaunchMethod


# ==============================================================================
#
class MPIExec(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, logger):

        LaunchMethod.__init__(self, cfg, logger)


    # --------------------------------------------------------------------------
    #
    def _configure(self):
        # mpiexec (e.g. on SuperMUC)
        self.launch_command = self._find_executable([
            'mpiexec',            # General case
            'mpiexec-mpich-mp',   # Mac OSX MacPorts
            'mpiexec-openmpi-mp'  # Mac OSX MacPorts
        ])

    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        opaque_slots = cu['opaque_slots']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_cores   = cud['cores']
        task_args    = cud.get('arguments') or []
        task_argstr  = self._create_arg_string(task_args)

        if not 'task_slots' in opaque_slots:
            raise RuntimeError('insufficient information to launch via %s: %s' \
                    % (self.name, opaque_slots))

        task_slots = opaque_slots['task_slots']

        # Extract all the hosts from the slots
        all_hosts = [slot.split(':')[0] for slot in task_slots]

        # Shorten the host list as much as possible
        hosts = self._compress_hostlist(all_hosts)

        # If we have a CU with many cores, and the compression didn't work
        # out, we will create a hostfile and pass  that as an argument
        # instead of the individual hosts
        if len(hosts) > 42:

            # Create a hostfile from the list of hosts
            hostfile = self._create_hostfile(all_hosts, separator=':')
            hosts_string = "-hostfile %s" % hostfile

        else:

            # Construct the hosts_string ('h1 h2 .. hN')
            hosts_string = "-host "+ ",".join(hosts)

        # Construct the executable and arguments
        if task_argstr:
            task_command = "%s %s" % (task_exec, task_argstr)
        else:
            task_command = task_exec

        mpiexec_command = "%s -n %s %s %s" % (
            self.launch_command, task_cores, hosts_string, task_command)

        return mpiexec_command, None


# ------------------------------------------------------------------------------




""" (Compute) Unit tests
"""
import os
import sys
import radical.pilot as rp
import unittest

import uuid
from copy import deepcopy
from radical.pilot.db import Session
from pymongo import MongoClient

# DBURL defines the MongoDB server URL and has the format mongodb://host:port.
# For the installation of a MongoDB server, refer to the MongoDB website:
# http://docs.mongodb.org/manual/installation/
DBURL = os.getenv("RADICAL_PILOT_DBURL")
if DBURL is None:
    print "ERROR: RADICAL_PILOT_DBURL (MongoDB server URL) is not defined."
    sys.exit(1)
    
DBNAME = os.getenv("RADICAL_PILOT_TEST_DBNAME")
if DBNAME is None:
    print "ERROR: RADICAL_PILOT_TEST_DBNAME (MongoDB database name) is not defined."
    sys.exit(1)


#-----------------------------------------------------------------------------
#
class TestIssue114(unittest.TestCase):
    # silence deprecation warnings under py3

    def setUp(self):
        # clean up fragments from previous tests
        client = MongoClient(DBURL)
        client.drop_database(DBNAME)

    def tearDown(self):
        # clean up after ourselves 
        client = MongoClient(DBURL)
        client.drop_database(DBNAME)

    def failUnless(self, expr):
        # St00pid speling.
        return self.assertTrue(expr)

    def failIf(self, expr):
        # St00pid speling.
        return self.assertFalse(expr)

    #-------------------------------------------------------------------------
    #
    def test__issue_114_part_1(self):
        """ https://github.com/radical-cybertools/radical.pilot/issues/114
        """
        session = rp.Session(database_url=DBURL, database_name=DBNAME)

        pm = rp.PilotManager(session=session)

        cpd = rp.ComputePilotDescription()
        cpd.resource = "local.localhost"
        cpd.cores    = 1
        cpd.runtime  = 5
        cpd.sandbox  = "/tmp/radical.pilot.sandbox.unittests"
        cpd.cleanup  = True

        pilot = pm.submit_pilots(pilot_descriptions=cpd)
        state = pm.wait_pilots(state=[rp.ACTIVE, 
                                      rp.DONE, 
                                      rp.FAILED], 
                                      timeout=5*60)

        assert (pilot.state == rp.ACTIVE), "pilot state: %s" % pilot.state

        um = rp.UnitManager(
            session=session,
            scheduler=rp.SCHED_DIRECT_SUBMISSION
        )
        um.add_pilots(pilot)

        all_tasks = []

        for i in range(0,2):
            cudesc = rp.ComputeUnitDescription()
            cudesc.cores      = 1
            cudesc.executable = "/bin/sleep"
            cudesc.arguments  = ['60']
            all_tasks.append(cudesc)

        units  = um.submit_units(all_tasks)
        states = um.wait_units (state=[rp.SCHEDULING, rp.EXECUTING], 
                                timeout=2*60)

        assert rp.SCHEDULING in states, "states: %s" % states

        states = um.wait_units (state=[rp.EXECUTING, rp.DONE], 
                                timeout=1*60)

        assert rp.EXECUTING  in states, "states: %s" % states

        session.close()

    #-------------------------------------------------------------------------
    #
    def test__issue_114_part_2(self):
        """ https://github.com/radical-cybertools/radical.pilot/issues/114
        """
        session = rp.Session(database_url=DBURL, database_name=DBNAME)

        pm  = rp.PilotManager(session=session)

        cpd = rp.ComputePilotDescription()
        cpd.resource = "local.localhost"
        cpd.cores   = 1
        cpd.runtime = 5
        cpd.sandbox = "/tmp/radical.pilot.sandbox.unittests"
        cpd.cleanup = True

        pilot = pm.submit_pilots(pilot_descriptions=cpd)

        um = rp.UnitManager(
            session=session,
            scheduler=rp.SCHED_DIRECT_SUBMISSION
        )
        um.add_pilots(pilot)

        state = pm.wait_pilots(state=[rp.ACTIVE, 
                                      rp.DONE, 
                                      rp.FAILED], 
                                      timeout=5*60)

        assert (pilot.state == rp.ACTIVE), "pilot state: %s" % pilot.state

        cudesc = rp.ComputeUnitDescription()
        cudesc.cores      = 1
        cudesc.executable = "/bin/sleep"
        cudesc.arguments  = ['60']

        cu    = um.submit_units(cudesc)
        state = um.wait_units(state=[rp.EXECUTING], timeout=60)

        assert state    == [rp.EXECUTING], 'state   : %s' % state
        assert cu.state ==  rp.EXECUTING , 'cu state: %s' % cu.state

        state = um.wait_units(timeout=2*60)

        assert state    == [rp.DONE], 'state   : %s' % state    
        assert cu.state ==  rp.DONE , 'cu state: %s' % cu.state 

        session.close()

    #-------------------------------------------------------------------------
    #
    def test__issue_114_part_3(self):
        """ https://github.com/radical-cybertools/radical.pilot/issues/114
        """
        session = rp.Session(database_url=DBURL, database_name=DBNAME)

        pm = rp.PilotManager(session=session)

        cpd = rp.ComputePilotDescription()
        cpd.resource = "local.localhost"
        cpd.cores   = 1
        cpd.runtime = 1
        cpd.sandbox = "/tmp/radical.pilot.sandbox.unittests"
        cpd.cleanup = True

        pilot = pm.submit_pilots(pilot_descriptions=cpd)

        um = rp.UnitManager(
            session   = session,
            scheduler = rp.SCHED_DIRECT_SUBMISSION
        )
        um.add_pilots(pilot)

        state = pm.wait_pilots(state=[rp.ACTIVE, 
                                      rp.DONE, 
                                      rp.FAILED], 
                                      timeout=10*60)

        assert state       == [rp.ACTIVE], 'state      : %s' % state    
        assert pilot.state ==  rp.ACTIVE , 'pilot state: %s' % pilot.state 

        state = pm.wait_pilots(timeout=3*60)

        assert state       == [rp.DONE], 'state      : %s' % state        
        assert pilot.state ==  rp.DONE , 'pilot state: %s' % pilot.state  

        session.close()





__copyright__ = "Copyright 2013-2015, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import time
import saga
import thread
import threading

from ..states import * 
from ..utils  import logger
from ..utils  import timestamp
from ..staging_directives import CREATE_PARENTS

IDLE_TIME  = 1.0  # seconds to sleep after idle cycles

# ----------------------------------------------------------------------------
#
class OutputFileTransferWorker(threading.Thread):
    """OutputFileTransferWorker handles the staging of the output files
    for a UnitManagerController.
    """

    # ------------------------------------------------------------------------
    #
    def __init__(self, session, unit_manager_id, number=None):

        self._session = session

        # threading stuff
        threading.Thread.__init__(self)
        self.daemon = True

        self.unit_manager_id = unit_manager_id

        self._worker_number = number
        self.name = "OutputFileTransferWorker-%s" % str(self._worker_number)

        # Stop event can be set to terminate the main loop
        self._stop = threading.Event()
        self._stop.clear()


    # ------------------------------------------------------------------------
    #
    def stop(self):
        """stop() signals the process to finish up and terminate.
        """
        logger.debug("otransfer %s stopping" % (self.name))
        self._stop.set()
        self.join()
        logger.debug("otransfer %s stopped" % (self.name))

    # ------------------------------------------------------------------------
    #
    def run(self):
        """Starts the process when Process.start() is called.
        """

        # make sure to catch sys.exit (which raises SystemExit)
        try :

            # Try to connect to the database and create a tailable cursor.
            try:
                db = self._session.get_db()
                um_col = db["%s.cu" % self._session.uid]
                logger.debug("Connected to MongoDB. Serving requests for UnitManager %s." % self.unit_manager_id)

            except Exception as e:
                logger.exception("Connection error: %s" % e)
                return

            while not self._stop.is_set() and \
                  not self._session._terminate.is_set():

                # See if we can find a ComputeUnit that is waiting for client output file transfer.
                # FIXME: this method is not bulkable.  See agent pulling for
                #        units for an approach to split the call into two bulkable 
                #        ones.
                ts = timestamp()
                compute_unit = um_col.find_and_modify(
                    query={"unitmanager": self.unit_manager_id,
                           "state"      : PENDING_OUTPUT_STAGING,
                           "control"    : 'agent'},
                    update={"$set" : {"state"   : STAGING_OUTPUT,
                                      "control" : 'umgr'},
                            "$push": {"statehistory" : 
                                         {"state"    : STAGING_OUTPUT, 
                                          "timestamp": ts}
                                     }
                            })

                if compute_unit is None:
                    # Sleep a bit if no new units are available.
                    time.sleep(IDLE_TIME)
                else:
                    logger.info("OFTW CU found, progressing ...")
                    state = STAGING_OUTPUT
                    compute_unit_id = None
                    try:
                        log_messages = []

                        # We have found a new CU. Now we can process the transfer
                        # directive(s) with SAGA.
                        compute_unit_id = str(compute_unit["_id"])

                        self._session.prof.prof('advance', uid=compute_unit_id,
                                msg=STAGING_OUTPUT, state=STAGING_OUTPUT)
                        logger.debug ("OutputStagingController: unit found: %s" % compute_unit_id)

                        remote_sandbox = compute_unit["sandbox"]
                        output_staging = compute_unit.get("FTW_Output_Directives", [])

                        logger.info("OutputStagingController: Processing output file transfers for ComputeUnit %s" % compute_unit_id)
                        # Loop over all staging directives and execute them.
                        for sd in output_staging:

                            logger.debug("OutputStagingController: sd: %s : %s" % (compute_unit_id, sd))

                            # Check if there was a cancel request for this CU
                            # TODO: Can't these cancel requests come from a central place?
                            state_doc = um_col.find_one(
                                {"_id": compute_unit_id},
                                fields=["state"]
                            )
                            if state_doc['state'] == CANCELED:
                                logger.info("Compute Unit Canceled, interrupting output file transfers.")
                                self._session.prof.prof('advance', uid=compute_unit_id, 
                                        msg=CANCELED, state=CANCELED)
                                state = CANCELED
                                # Break out of the loop over all SD's, into the loop over CUs
                                break

                            abs_src = "%s/%s" % (remote_sandbox, sd['source'])

                            if os.path.basename(sd['target']) == sd['target']:
                                abs_target = "file://localhost%s" % os.path.join(os.getcwd(), sd['target'])
                            else:
                                abs_target = "file://localhost%s" % os.path.abspath(sd['target'])

                            log_msg = "Transferring output file %s -> %s" % (abs_src, abs_target)
                            log_messages.append(log_msg)
                            logger.debug(log_msg)

                            output_file = saga.filesystem.File(saga.Url(abs_src),
                                                               session=self._session)

                            if CREATE_PARENTS in sd['flags']:
                                copy_flags = saga.filesystem.CREATE_PARENTS
                            else:
                                copy_flags = 0

                            try:
                                output_file.copy(saga.Url(abs_target), flags=copy_flags)
                                output_file.close()
                            except Exception as e:
                                logger.exception(e)
                                raise Exception("output stagingfor %s failed: %s" \
                                                % (compute_unit_id, e))

                        # If the CU was canceled we can skip the remainder of this loop,
                        # and return to the CU loop
                        if state == CANCELED:
                            continue

                        # Update the CU's state to 'DONE'.
                        ts = timestamp()
                        log_message = "Output transfer completed."
                        um_col.update({'_id': compute_unit_id}, {
                            '$set': {'state': DONE},
                            '$push': {
                                'statehistory': {'state': DONE, 'timestamp': ts},
                                'log': {'message': log_message, 'timestamp': ts}
                            }
                        })
                        self._session.prof.prof('advance', uid=compute_unit_id,
                                msg=DONE, state=DONE)

                    except Exception as e :
                        # Update the CU's state to 'FAILED'.
                        ts = timestamp()
                        log_message = "Output transfer failed: %s" % e
                        um_col.update({'_id': compute_unit_id}, {
                            '$set': {'state': FAILED},
                            '$push': {
                                'statehistory': {'state': FAILED, 'timestamp': ts},
                                'log': {'message': log_message, 'timestamp': ts}
                            }})
                        logger.exception(log_message)
                        self._session.prof.prof('advance', uid=compute_unit_id, 
                                msg=FAILED, state=FAILED)
                        raise

        except SystemExit as e :
            logger.exception("output file transfer thread caught system exit -- forcing application shutdown")
            thread.interrupt_main()




__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"


# ------------------------------------------------------------------------------
# constants and types
from .types      import *
from .states     import *
from .logentry   import * 
from .scheduler  import *
from .constants  import *
from .exceptions import *


# ------------------------------------------------------------------------------
# import API
from .session                   import Session, default_dburl
from .context                   import Context

from .unit_manager              import UnitManager
from .compute_unit              import ComputeUnit
from .compute_unit_description  import ComputeUnitDescription

from .pilot_manager             import PilotManager
from .compute_pilot             import ComputePilot
from .compute_pilot_description import ComputePilotDescription

from .resource_config           import ResourceConfig
from .staging_directives        import COPY, LINK, MOVE, TRANSFER
from .staging_directives        import SKIP_FAILED, CREATE_PARENTS

from .utils                     import version, version_detail, version_branch
from .utils                     import sdist_name, sdist_path
from .utils                     import logger


# ------------------------------------------------------------------------------
# make submodules available -- mostly for internal use
import utils
import worker
import agent


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import sys
import radical.pilot
import radical.utils as ru
import os


dh = ru.DebugHelper ()

RUNTIME         =   10
CORES           =   16
SIM_INSTANCES   =   16
ANA_INSTANCES   =   1           # DO NOT CHANGE !
CYCLES          =   2
SCHED       = radical.pilot.SCHED_DIRECT_SUBMISSION

MY_STAGING_AREA = 'staging:///'

resources = {
        'local.localhost' : {
            'project'  : None,
            'queue'    : None,
            'schema'   : None
            },

        'epsrc.archer' : {
            'project'  : 'e290',
            'queue'    : 'short',
            'schema'   : None
            },

        'xsede.stampede' : {
            'project'  : 'TG-MCB090174',
            'queue'    : 'development',
            'schema'   : None
            },

        }

remote_env = {
    

    'amber': {

        'local.localhost': [],      #For the user to fill in order to run on localhost

        'epsrc.archer': [
                            "module load packages-archer",
                            "module load amber"
                        ],

        'xsede.stampede': [
                            "module load TACC", 
                            "module load amber/12.0"
                        ],

    },

    'coco': {

        'local.localhost': [],      #For the user to fill in order to run on localhost

        'epsrc.archer': [

                           "module load python-compute/2.7.6",
                            "module load pc-numpy/1.8.0-libsci",
                            "module load pc-scipy/0.13.3-libsci",
                            "module load pc-coco/0.18",
                            "module load pc-netcdf4-python/1.1.0",
                            "module load amber"
                        ],

            
        'xsede.stampede': [
                            "module load TACC",
                            "module load intel/13.0.2.146",
                            "module load python/2.7.9",
                            "module load netcdf/4.3.2",
                            "module load hdf5/1.8.13",
                            "export PYTHONPATH=/opt/apps/intel13/mvapich2_1_9/python/2.7.9/lib/python2.7/site-packages:/work/02998/ardi/coco-0.19_installation/lib/python2.7/site-packages:$PYTHONPATH",
                            "export PATH=/work/02998/ardi/coco-0.19_installation/bin:$PATH",
                            "module load amber"
                        ],

    },
}

#------------------------------------------------------------------------------
#
def pilot_state_cb (pilot, state):

    if pilot:

        print "[Callback]: ComputePilot '{0}' state changed to {1}.".format(
            pilot.uid, state)

        if state == radical.pilot.states.FAILED:
            print "#######################"
            print "##       ERROR       ##"
            print "#######################"
            print "Pilot {0} has FAILED. Can't recover.".format(pilot.uid)
            print "Pilot log:- "
            for log in pilot.log:
                print log.as_dict()
            print u'Pilot STDOUT : {0}'.format(pilot.stdout)
            print u'Pilot STDERR : {0}'.format(pilot.stderr)
            sys.exit(1)


#------------------------------------------------------------------------------
#
def unit_state_cb (unit, state):

    
    if unit:
        print "[Callback]: ComputeUnit '{0}' state changed to {1}.".format(
            unit.uid, state)

        if state == radical.pilot.states.FAILED:
            print "#######################"
            print "##       ERROR       ##"
            print "#######################"
            print "ComputeUnit {0} has FAILED. Can't recover.".format(unit.uid)
            print "ComputeUnit log:- "
            for log in unit.log:
                print log.as_dict()
            print u"STDERR : {0}".format(unit.stderr)
            print u"STDOUT : {0}".format(unit.stdout)
            sys.exit(1)

        elif state == radical.pilot.states.CANCELED:
            print "#######################"
            print "##       ERROR       ##"
            print "#######################"
            print "ComputeUnit {0} was canceled prematurely because the pilot was terminated. Can't recover.".format(unit.uid)
            sys.exit(1)

    else:
        return


#------------------------------------------------------------------------------
#
def wait_queue_size_cb(umgr, wait_queue_size):

    print "[Callback]: wait_queue_size: %s." % wait_queue_size


#------------------------------------------------------------------------------
#
if __name__ == "__main__":

    # we can optionally pass session name to RP
    if len(sys.argv) > 1:
        resource = sys.argv[1]
    else:
        resource = 'local.localhost'

    print 'running on %s' % resource

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = radical.pilot.Session()
    # Add an ssh identity to the session.
    cred = radical.pilot.Context('ssh')
    cred.user_id = 'vivek91'
    session.add_context(cred)

    print "session id: %s" % session.uid

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        pmgr = radical.pilot.PilotManager(session=session)
        pmgr.register_callback(pilot_state_cb)

        pdesc = radical.pilot.ComputePilotDescription()
        pdesc.resource      = resource
        pdesc.cores         = CORES
        pdesc.project       = resources[resource]['project']
        pdesc.queue         = resources[resource]['queue']
        pdesc.runtime       = RUNTIME
        pdesc.cleanup       = False
        pdesc.access_schema = resources[resource]['schema']


        pilot = pmgr.submit_pilots(pdesc)


        #------------------------------------------------------------------------------------
        # Data staging with pilot
        pilot_stagein_files = [
                                'mdshort.in',
                                'min.in',
                                'penta.top',
                                'penta.crd',
                                'postexec.py'
                            ]

        md_file, min_file, top_file, crd_file, postexec_file = pilot_stagein_files

        sd_pilot_list = list()
        for item in pilot_stagein_files:
            sd_pilot = {
                        'source': 'file://{0}/{1}'.format(os.getcwd(),item),
                        'target': '{0}{1}'.format(MY_STAGING_AREA,item),
                        'action': radical.pilot.TRANSFER
                        }
            sd_pilot_list.append(sd_pilot)

        sd_pilot = {
                        'source': 'file://{0}/{1}'.format(os.getcwd(),crd_file),
                        'target': '{0}iter0/{1}'.format(MY_STAGING_AREA,crd_file),
                        'action': radical.pilot.TRANSFER
                    }
        sd_pilot_list.append(sd_pilot)

        pilot.stage_in(sd_pilot_list)

        #------------------------------------------------------------------------------------

        umgr = radical.pilot.UnitManager(session=session, scheduler=SCHED)
        umgr.register_callback(unit_state_cb,      radical.pilot.UNIT_STATE)
        umgr.register_callback(wait_queue_size_cb, radical.pilot.WAIT_QUEUE_SIZE)
        umgr.add_pilots(pilot)

        cycle=0
        
        for cycle in range(0,CYCLES):

            print 'Cycle %s'%(cycle+1)
            print '======================'

            print 'Simulation Step'
            print '----------------------'
            # Minimization Step
            cudesc_list_A = list()
            for i in range(0, SIM_INSTANCES):

                #==================================================================
                # CU Definition for Amber - Stage 1

                cud = radical.pilot.ComputeUnitDescription()
                cud.executable     = 'pmemd'
                cud.pre_exec       = remote_env['amber'][resource]
                cud.arguments      = [
                                        '-O',
                                        '-i',min_file,
                                        '-o','min%s.out'%cycle,
                                        '-inf','min%s.inf'%cycle,
                                        '-r','md%s.crd'%cycle,
                                        '-p',top_file,
                                        '-c',crd_file,
                                        '-ref','min%s.crd'%cycle
                                    ]
                cud.cores          = 1

                #==================================================================
                # Data Staging for the CU

                #------------------------------------------------------------------
                # Coordinate file staging

                if(cycle==0):

                    crd_stage = {
                                'source': MY_STAGING_AREA + 'iter0/' + crd_file,
                                'target': 'min0.crd',
                                'action': radical.pilot.LINK
                                }
                else:

                    crd_stage = {
                                'source': MY_STAGING_AREA + 'iter{0}/min{0}{1}.crd'.format(cycle,i),
                                'target': 'min{0}.crd'.format(cycle),
                                'action': radical.pilot.LINK
                                }

                #------------------------------------------------------------------

                #------------------------------------------------------------------
                # Configure the staging directive for shared input file.
                top_stage = {
                            'source': MY_STAGING_AREA + top_file,
                            'target': top_file,
                            'action': radical.pilot.LINK
                            }
                minin_stage = {
                            'source': MY_STAGING_AREA + min_file,
                            'target': min_file,
                            'action': radical.pilot.LINK
                            }

                init_crd_stage = {
                            'source': MY_STAGING_AREA + crd_file,
                            'target': crd_file,
                            'action': radical.pilot.LINK
                            }
                #------------------------------------------------------------------

                #------------------------------------------------------------------
                # Stage OUT the output to the staging area

                md_stage_out = {
                            'source': 'md{0}.crd'.format(cycle),
                            'target': MY_STAGING_AREA + 'iter{0}/md{0}{1}.crd'.format(cycle,i),
                            'action': radical.pilot.LINK
                            }
                #----------------------------------------------------------

                cud.input_staging = [crd_stage,top_stage,minin_stage,init_crd_stage]
                cud.output_staging = [md_stage_out]

                cudesc_list_A.append(cud)


            # MD step
            cudesc_list_B = []
            for i in range(0,SIM_INSTANCES):

                #==================================================================
                # CU Definition for Amber - Stage 2

                cud = radical.pilot.ComputeUnitDescription()
                cud.executable     = 'pmemd'
                cud.pre_exec       = remote_env['amber'][resource]
                cud.arguments      = [
                                    '-O',
                                    '-i', md_file,
                                    '-o', 'md%s.out'%cycle,
                                    '-inf', 'md%s.inf'%cycle,
                                    '-x', 'md%s.ncdf'%cycle,
                                    '-r', 'md%s.rst'%cycle,
                                    '-p', top_file,
                                    '-c', 'md%s.crd'%cycle
                                    ]
                cud.cores          = 1

                #==================================================================
                # Data Staging

                #------------------------------------------------------------------
                # Link to output from first-stage of Amber
                md_stage_in = {
                                'source': MY_STAGING_AREA + 'iter{0}/md{0}{1}.crd'.format(cycle,i),
                                'target': 'md{0}.crd'.format(cycle),
                                'action': radical.pilot.LINK
                                }

                # Link to shared data from staging area
                mdin_stage = {
                                'source': MY_STAGING_AREA + md_file,
                                'target': md_file,
                                'action': radical.pilot.LINK
                            }
                top_stage = {
                                'source': MY_STAGING_AREA + top_file,
                                'target': top_file,
                                'action': radical.pilot.LINK
                            }

                cud.input_staging = [md_stage_in,mdin_stage,top_stage]
                #------------------------------------------------------------------


                #------------------------------------------------------------------
                # Link output data to staging area
                ncdf_stage_out = {
                                'source': 'md{0}.ncdf'.format(cycle),
                                'target': MY_STAGING_AREA + 'iter{0}/md_{0}_{1}.ncdf'.format(cycle, i),
                                'action': radical.pilot.COPY
                                }
                cud.output_staging = [ncdf_stage_out]
                #------------------------------------------------------------------

                cudesc_list_B.append(cud) 


            #Submit all Amber CUs in Stage 1
            cu_list_A = umgr.submit_units(cudesc_list_A)

            cu_list_B = []
            cu_list_A_copy = cu_list_A[:]

            while cu_list_A:
                for cu_a in cu_list_A:
                    idx = cu_list_A_copy.index(cu_a)
                    cu_a.wait ()
                    cu_list_B.append(umgr.submit_units(cudesc_list_B[idx]))
                    cu_list_A.remove(cu_a)

            umgr.wait_units()    


            #==================================================================
            # CU Definition for CoCo

            print 'Analysis Step'
            print '----------------------'

            cud = radical.pilot.ComputeUnitDescription()
            cud.cores = SIM_INSTANCES  #pyCoCo should use as many as cores as the number of frontpoints
            cud.executable = 'pyCoCo'
            cud.pre_exec = remote_env['coco'][resource]
            cud.arguments = [
                                '--grid','5',
                                '--dims','3',
                                '--frontpoints',SIM_INSTANCES,
                                '--topfile',top_file,
                                '--mdfile','*.ncdf',
                                '--output','pdbs',
                                '--logfile','coco.log',
                            ]
            cud.post_exec = ['python postexec.py %s %s' % (SIM_INSTANCES,cycle)]
            cud.mpi = False


            #==================================================================
            # Data Staging for the CU

            #------------------------------------------------------------------
            # postexec and topology file staging

            postexec_stage = {
                            'source': MY_STAGING_AREA + 'postexec.py',
                            'target': 'postexec.py',
                            'action': radical.pilot.LINK
                        }

            top_stage = {
                            'source': MY_STAGING_AREA + top_file,
                            'target': top_file,
                            'action': radical.pilot.LINK
                        }
            cud.input_staging = [postexec_stage,top_stage]
            #------------------------------------------------------------------


            #------------------------------------------------------------------
            # stage in the ncdf files from all previous iterations
            for iter in range(0,cycle+1):
                for inst in range(0,SIM_INSTANCES):
                    dir = {
                            'source': MY_STAGING_AREA + 'iter{0}/md_{0}_{1}.ncdf'.format(iter,inst),
                            'target': 'md_{0}_{1}.ncdf'.format(iter,inst),
                            'action': radical.pilot.COPY
                            }
                    cud.input_staging.append(dir)
            #------------------------------------------------------------------

            cud.output_staging = []
            #------------------------------------------------------------------
            # stage out the crd files to staging area
            for inst in range(0,SIM_INSTANCES):
                dir = {
                        'source': 'min{0}{1}.crd'.format(cycle,inst),
                        'target': MY_STAGING_AREA + 'iter{2}/min{0}{1}.crd'.format(cycle,inst,cycle+1),
                        'action': radical.pilot.LINK
                        }
                cud.output_staging.append(dir)
            #------------------------------------------------------------------

            #------------------------------------------------------------------
            # stage out the crd files to localhost
            for inst in range(0,SIM_INSTANCES):
                dir = {
                        'source': 'min{0}{1}.crd'.format(cycle,inst),
                        'target': 'backup/iter{1}/min{0}.crd'.format(inst,cycle+1),
                        }
                cud.output_staging.append(dir)
            #------------------------------------------------------------------

            unit = umgr.submit_units(cud)

            unit.wait()



    except Exception as e:
        # Something unexpected happened in the pilot code above
        print "caught Exception: %s" % e
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        print "need to exit now: %s" % e

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        print "closing session"
        session.close ()

        # the above is equivalent to
        #
        #   session.close (cleanup=True, terminate=True)
        #
        # it will thus both clean out the session's database record, and kill
        # all remaining pilots (none in our example).


#-------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import shutil

import radical.utils as ru

from .... import pilot     as rp
from ...  import utils     as rpu
from ...  import states    as rps
from ...  import constants as rpc

from .base import AgentStagingOutputComponent


# ==============================================================================
#
class Default(AgentStagingOutputComponent):
    """
    This component performs all agent side output staging directives for compute
    units.  It gets units from the agent_staging_output_queue, in
    AGENT_STAGING_OUTPUT_PENDING state, will advance them to
    AGENT_STAGING_OUTPUT state while performing the staging, and then moves then
    to the UMGR_STAGING_OUTPUT_PENDING state, which at the moment requires the
    state change to be published to MongoDB (no push into a queue).

    Note that this component also collects stdout/stderr of the units (which
    can also be considered staging, really).
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg):

        rpu.Component.__init__(self, 'AgentStagingOutputComponent', cfg)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def create(cls, cfg):

        return cls(cfg)


    # --------------------------------------------------------------------------
    #
    def initialize_child(self):

        self.declare_input (rps.AGENT_STAGING_OUTPUT_PENDING, rpc.AGENT_STAGING_OUTPUT_QUEUE)
        self.declare_worker(rps.AGENT_STAGING_OUTPUT_PENDING, self.work)

        # we don't need an output queue -- units are picked up via mongodb
        self.declare_output(rps.PENDING_OUTPUT_STAGING, None) # drop units

        self.declare_publisher('state', rpc.AGENT_STATE_PUBSUB)

        # all components use the command channel for control messages
        self.declare_publisher ('command', rpc.AGENT_COMMAND_PUBSUB)

        # communicate successful startup
        self.publish('command', {'cmd' : 'alive',
                                 'arg' : self.cname})


    # --------------------------------------------------------------------------
    #
    def finalize_child(self):

        # communicate finalization
        self.publish('command', {'cmd' : 'final',
                                 'arg' : self.cname})


    # --------------------------------------------------------------------------
    #
    def work(self, cu):

        self.advance(cu, rps.AGENT_STAGING_OUTPUT, publish=True, push=False)

        staging_area = os.path.join(self._cfg['workdir'], self._cfg['staging_area'])
        staging_ok   = True

        workdir = cu['workdir']

        ## parked from unit state checker: unit postprocessing
        if os.path.isfile(cu['stdout_file']):
            with open(cu['stdout_file'], 'r') as stdout_f:
                try:
                    txt = unicode(stdout_f.read(), "utf-8")
                except UnicodeDecodeError:
                    txt = "unit stdout contains binary data -- use file staging directives"

                cu['stdout'] += rpu.tail(txt)

        if os.path.isfile(cu['stderr_file']):
            with open(cu['stderr_file'], 'r') as stderr_f:
                try:
                    txt = unicode(stderr_f.read(), "utf-8")
                except UnicodeDecodeError:
                    txt = "unit stderr contains binary data -- use file staging directives"

                cu['stderr'] += rpu.tail(txt)

        if 'RADICAL_PILOT_PROFILE' in os.environ:
            if os.path.isfile("%s/PROF" % cu['workdir']):
                try:
                    with open("%s/PROF" % cu['workdir'], 'r') as prof_f:
                        txt = prof_f.read()
                        for line in txt.split("\n"):
                            if line:
                                x1, x2, x3 = line.split()
                                self._prof.prof(x1, msg=x2, timestamp=float(x3), uid=cu['_id'])
                except Exception as e:
                    self._log.error("Pre/Post profiling file read failed: `%s`" % e)

        # NOTE: all units get here after execution, even those which did not
        #       finish successfully.  We do that so that we can make
        #       stdout/stderr available for failed units.  But at this point we
        #       don't need to advance those units anymore, but can make them
        #       final.
        if cu['target_state'] != rps.DONE:
            self.advance(cu, cu['target_state'], publish=True, push=False)
            return


        try:
            # all other units get their (expectedly valid) output files staged
            for directive in cu['Agent_Output_Directives']:

                self._prof.prof('Agent output_staging', uid=cu['_id'],
                         msg="%s -> %s" % (str(directive['source']), str(directive['target'])))

                # Perform output staging
                self._log.info("unit output staging directives %s for cu: %s to %s",
                        directive, cu['_id'], workdir)

                # Convert the target_url into a RU Url object
                target_url = ru.Url(directive['target'])

                # Handle special 'staging' scheme
                if target_url.scheme == self._cfg['staging_scheme']:
                    self._log.info('Operating from staging')
                    # Remove the leading slash to get a relative path from
                    # the staging area
                    rel2staging = target_url.path.split('/',1)[1]
                    target = os.path.join(staging_area, rel2staging)
                else:
                    self._log.info('Operating from absolute path')
                    # FIXME: will this work for TRANSFER mode?
                    target = target_url.path

                # Get the source from the directive and convert it to the location
                # in the workdir
                source = str(directive['source'])
                abs_source = os.path.join(workdir, source)

                # Create output directory in case it doesn't exist yet
                # FIXME: will this work for TRANSFER mode?
                rpu.rec_makedir(os.path.dirname(target))

                self._log.info("Going to '%s' %s to %s", directive['action'], abs_source, target)

                if directive['action'] == rpc.LINK:
                    # This is probably not a brilliant idea, so at least give a warning
                    os.symlink(abs_source, target)
                elif directive['action'] == rpc.COPY:
                    shutil.copyfile(abs_source, target)
                elif directive['action'] == rpc.MOVE:
                    shutil.move(abs_source, target)
                else:
                    # FIXME: implement TRANSFER mode
                    raise NotImplementedError('Action %s not supported' % directive['action'])

                log_message = "%s'ed %s to %s - success" %(directive['action'], abs_source, target)
                self._log.info(log_message)

        except Exception as e:
            self._log.exception("staging output failed -> unit failed")
            staging_ok = False


        # Agent output staging is done (or failed)
        if staging_ok:
          # self.advance(cu, rps.UMGR_STAGING_OUTPUT_PENDING, publish=True, push=True)
            self.advance(cu, rps.PENDING_OUTPUT_STAGING, publish=True, push=False)
        else:
            self.advance(cu, rps.FAILED, publish=True, push=False)



# ------------------------------------------------------------------------------





import os

from ..states import *

info_names = {
        'AgentWorker'                 : 'awo',
        'AgentStagingInputComponent'  : 'asic',
        'SchedulerContinuous'         : 'asc',  # agent scheduler component
        'AgentExecutingComponent'     : 'aec',
        'AgentStagingOutputComponent' : 'asoc',
        'session'                     : 'mod'
        }

_info_events = {
        'get'        : '_get_u',     # get a unit from a queue
        'work start' : '_work_u',    # unit is handed over to component
        'work done'  : '_worked_u',  # component finished to operate on unit
        'put'        : '_put_u',     # unit is put onto the next queue
        'publish'    : '_pub_u',     # unit state is published via some pubsub
        'advance'    : '_adv_u',     # the unit state is advanced
        'update'     : '_upd_u'      # a unit state update is pushed to the DB
        }

_info_pending = {
        'Pending'    : '_pend'
        }

_info_premature_final = {
        'Failed'   : '_fail',
        'Canceled'   : '_canc'
}

_info_states = [
        ACTIVE,
        AGENT_STAGING_INPUT,
        AGENT_STAGING_INPUT_PENDING,
        AGENT_STAGING_OUTPUT,
        AGENT_STAGING_OUTPUT_PENDING,
        ALLOCATING,
        ALLOCATING_PENDING,
        CANCELED,
        DONE,
        EXECUTING,
        EXECUTING_PENDING,
        FAILED,
        LAUNCHING,
        NEW,
        PENDING,
        PENDING_ACTIVE,
        PENDING_EXECUTION,
        PENDING_INPUT_STAGING,
        PENDING_LAUNCH,
        PENDING_OUTPUT_STAGING,
        SCHEDULING,
        STAGING_INPUT,
        STAGING_OUTPUT,
        UNSCHEDULED
        ]

_info_entries = [
    ('umgr_get_u',      'MainThread',             'advance',   'New'),
    ('umgr_adv_u_pend', 'MainThread',             'advance',   'PendingInputStaging'),
    ('usic_get_u',      'InputFileTransfer',      'advance',   'StagingInput'),
    ('usic_adv_u_pend', 'InputFileTransfer',      'advance',   'AgentStagingInputPending'),

    ('usoc_get_u',      'OutputFileTransfer',     'advance',   'StagingOutput'),
    ('usoc_adv_u',      'OutputFileTransfer',     'advance',   'Done'),

    # FIXME: the names below will break for other schedulers
    ('asc_allocated',   'SchedulerContinuous',    'schedule',  'allocated'),
    ('asc_alloc_nok',   'SchedulerContinuous',    'schedule',  'allocation failed'),
    ('asc_alloc_ok',    'SchedulerContinuous',    'schedule',  'allocation succeeded'),
    ('asc_unqueue',     'SchedulerContinuous',    'unqueue',   're-allocation done'),
    ('asc_released',    'SchedulerContinuous',    'unschedule','released'),

    ('aec_launch',      'AgentExecuting',         'exec',      'unit launch'),
    ('aec_spawn',       'AgentExecuting',         'spawn',     'unit spawn'),
    ('aec_script',      'AgentExecuting',         'command',   'launch script constructed'),
    ('ace_outerr',      'AgentExecuting',         'command',   'stdout and stderr files created'),
    ('aec_handover',    'AgentExecuting',         'spawn',     'spawning passed to pty'),
    ('aec_handover',    'AgentExecuting',         'spawn',     'spawning passed to popen'),
    ('aec_end',         'AgentExecuting',         'final',     ''),

    ('aec_pickup',      'AgentExecuting',         'passed',    'ExecWatcher picked up unit'),
    ('aec_start_script','AgentStagingOutputComponent','script','start_script'),
    ('aec_after_cd',    'AgentStagingOutputComponent','script','after_cd'),
    ('aec_after_exec',  'AgentStagingOutputComponent','script','after_exec'),
    ('aec_complete',    'AgentExecuting',         'exec',      'execution complete'),
]

# ------------------------------------------------------------------------------
#
tmp = None
def add_concurrency (frame, tgt, spec):
    """
    add a column 'tgt' which is a cumulative sum of conditionals of another row.
    
    The purpose is the following: if a unit enters a component, the tgt row counter is 
    increased by 1, if the unit leaves the component, the counter is decreased by 1.
    For any time, the resulting row contains the number of units which is in the 
    component.  Or state.  Or whatever.
    
    The arguments are:
        'tgt'  : name of the new column
        'spec' : a set of filters to determine if a unit enters or leaves
    
    'spec' is expected to be a dict of the following format:
    
        spec = { 'in'  : [{'col1' : 'pat1', 
                           'col2' : 'pat2'},
                          ...],
                 'out' : [{'col3' : 'pat3', 
                           'col4' : 'pat4'},
                          ...]
               }
    
    where:
        'in'    : filter set to determine the unit entering
        'out'   : filter set to determine the unit leaving
        'col'   : name of column for which filter is defined
        'event' : event which correlates to entering/leaving
        'msg'   : qualifier on the event, if event is not unique
    
    Example:
        spec = {'in'  : [{'state' :'Executing'}],
                'out' : [{'state' :'Done'},
                         {'state' :'Failed'},
                         {'state' :'Canceled'}]
               }
        add_concurrency (df, 'concurrently_running', spec)
    """
    
    import numpy as np

    # create a temporary row over which we can do the commulative sum
    # --------------------------------------------------------------------------
    def _conc (row, spec):

        # row must match any filter dict in 'spec[in/out]' 
        # for any filter dict it must match all col/pat pairs

        # for each in filter
        for f in spec['in']:
            match = 1 
            # for each col/val in that filter
            for col, pat in f.iteritems():
                if row[col] != pat:
                    match = 0
                    break
            if match:
                # one filter matched!
              # print " + : %-20s : %.2f : %-20s : %s " % (row['uid'], row['time'], row['event'], row['message'])
                return 1

        # for each out filter
        for f in spec['out']:
            match = 1 
            # for each col/val in that filter
            for col, pat in f.iteritems():
                if row[col] != pat:
                    match = 0
                    break
            if match:
                # one filter matched!
              # print " - : %-20s : %.2f : %-20s : %s " % (row['uid'], row['time'], row['event'], row['message'])
                return -1

        # no filter matched
      # print "   : %-20s : %.2f : %-20s : %s " % (row['uid'], row['time'], row['event'], row['message'])
        return  np.NaN
    # --------------------------------------------------------------------------

    # we only want to later look at changes of the concurrency -- leading or trailing 
    # idle times are to be ignored.  We thus set repeating values of the cumsum to NaN, 
    # so that they can be filtered out when ploting: df.dropna().plot(...).  
    # That specifically will limit the plotted time range to the area of activity. 
    # The full time range can still be plotted when ommitting the dropna() call.
    # --------------------------------------------------------------------------
    def _time (x):
        global tmp
        if     x != tmp: tmp = x
        else           : x   = np.NaN
        return x


    # --------------------------------------------------------------------------
    # sanitize concurrency: negative values indicate incorrect event ordering,
    # so we set the repesctive values to 0
    # --------------------------------------------------------------------------
    def _abs (x):
        if x < 0:
            return np.NaN
        return x
    # --------------------------------------------------------------------------
    
    frame[tgt] = frame.apply(lambda row: _conc(row, spec), axis=1).cumsum()
    frame[tgt] = frame.apply(lambda row: _abs (row[tgt]),  axis=1)
    frame[tgt] = frame.apply(lambda row: _time(row[tgt]),  axis=1)
  # print frame[[tgt, 'time']]

    return frame



# ------------------------------------------------------------------------------
#
def add_frequency(frame, tgt, window, spec):
    """
    This method will add a row 'tgt' to the given data frame, which will contain
    a contain the frequency (1/s) of the events specified in 'spec'.

    We first will filter the given frame by spec, and then apply a rolling
    window over the time column, counting the rows which fall into the window.
    The result is *not* divided by window size, so normalization is up to the
    caller.
    
    The method looks backwards, so the resulting frequency column contains the
    frequency which applied *up to* that point in time.
    """
    
    # --------------------------------------------------------------------------
    def _freq(t, _tmp, _window):
        # get sequence of frame which falls within the time window, and return
        # length of that sequence
        return len(_tmp.uid[(_tmp.time > t-_window) & (_tmp.time <= t)])
    # --------------------------------------------------------------------------
    
    # filter the frame by the given spec
    tmp = frame
    for key,val in spec.iteritems():
        tmp = tmp[tmp[key].isin([val])]
    frame[tgt] = tmp.time.apply(_freq, args=[tmp, window])

    return frame


# ------------------------------------------------------------------------------
#
def add_event_count(frame, tgt, spec):
    """
    This method will add a row 'tgt' to the given data frame, which will contain
    a counter of the events specified in 'spec'.

    This works similar to add_frequency: we first filter, and then add the
    cumsum. 
    """

    raise NotImplementedError('not yet implemented')

    # --------------------------------------------------------------------------
    def _ecnt(t, _tmp):
        # get sequence of frame which falls within the time window, and return
        # length of that sequence
        return len(_tmp.uid[(_tmp.time <= t)])
    # --------------------------------------------------------------------------

    # filter the frame by the given spec
    tmp = frame
    for key,val in spec.iteritems():
        tmp = tmp[tmp[key].isin([val])]
    frame[tgt] = tmp.time.apply(_ecnt, args=[tmp])

    return frame


# ------------------------------------------------------------------------------
#
t0 = None
def calibrate_frame(frame, spec):
    """
    move the time axis of a profiling frame so that t_0 is at the first event
    matching the given 'spec'.  'spec' has the same format as described in
    'add_concurrency' (list of dicts with col:pat filters)
    """

    # --------------------------------------------------------------------------
    def _find_t0 (row, spec):

        # row must match any filter dict in 'spec[in/out]' 
        # for any filter dict it must match all col/pat pairs
        global t0
        if t0 is not None:
            # already found t0
            return

        # for each col/val in that filter
        for f in spec:
            match = 1 
            for col, pat in f.iteritems():
                if row[col] != pat:
                    match = 0
                    break
            if match:
                # one filter matched!
                t0 = row['time']
                return
    # --------------------------------------------------------------------------

    # --------------------------------------------------------------------------
    def _calibrate (row, t0):

        if t0 is None:
            # no t0...
            return

        return row['time'] - t0
    # --------------------------------------------------------------------------

    # we need to iterate twice over the frame: first to find t0, then to
    # calibrate the time axis
    global t0
    t0 = None # no t0
    frame.apply(lambda row: _find_t0  (row, spec), axis=1)

    if t0 == None:
        print "Can't recalibrate, no matching timestamp found"
        return
    frame['time'] = frame.apply(lambda row: _calibrate(row, t0  ), axis=1)

    return frame


# ------------------------------------------------------------------------------
#
def create_plot():
    """
    create a plot object and tune its layout to our liking.
    """
    
    import matplotlib.pyplot as plt

    fig, plot = plt.subplots(figsize=(12,6))
    
    plot.xaxis.set_tick_params(width=1, length=7)
    plot.yaxis.set_tick_params(width=1, length=7)

    plot.spines['right' ].set_position(('outward', 10))
    plot.spines['top'   ].set_position(('outward', 10))
    plot.spines['bottom'].set_position(('outward', 10))
    plot.spines['left'  ].set_position(('outward', 10))

    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)
    
    fig.tight_layout()

    return fig, plot


# ------------------------------------------------------------------------------
#
def frame_plot (frames, axis, title=None, logx=False, logy=False, 
                legend=True, figdir=None):
    """
    plot the given axis from the give data frame.  We create a plot, and plot
    all frames given in the list.  The list is expected to contain [frame,label]
    pairs
    
    frames: list of tuples of dataframes and labels
    frames  = [[stampede_df_1, 'stampede - popen'], 
               [stampede_df_2, 'stampede - shell'],
               [stampede_df_3, 'stampede - ORTE' ]]
     
    axis:   tuple of data frame column index and axis label
    axis    = ['time', 'time (s)']
    """
    
    # create figure and layout
    fig, plot = create_plot()

    # set plot title
    if title:
        plot.set_title(title, y=1.05, fontsize=18)

    # plot the data frames
    # NOTE: we need to set labels separately, because of
    #       https://github.com/pydata/pandas/issues/9542
    labels = list()
    for frame, label in frames:
        try:
            frame.dropna().plot(ax=plot, logx=logx, logy=logy,
                    x=axis[0][0], y=axis[1][0],
                    drawstyle='steps',
                    label=label, legend=False)
        except Exception as e:
            print "skipping frame '%s': '%s'" % (label, e)

    if legend:
        plot.legend(labels=labels, loc='upper right', fontsize=14, frameon=True)

    # set axis labels
    plot.set_xlabel(axis[0][1], fontsize=14)
    plot.set_ylabel(axis[1][1], fontsize=14)
    plot.set_frame_on(True)
   
    # save as png and pdf.  Use the title as base for names
    if title: base = title
    else    : base = "%s_%s" % (axis[0][1], axis[1][1])
        
    # clean up base name -- only keep alphanum and such
    import re
    base = re.sub('[^a-zA-Z0-9\.\-]', '_', base)
    base = re.sub('_+',               '_', base)
    
    if not figdir:
        figdir = os.getcwd()

    print 'saving %s/%s.png' % (figdir, base)
    fig.savefig('%s/%s.png' % (figdir, base), bbox_inches='tight')

    print 'saving %s/%s.pdf' % (figdir, base)
    fig.savefig('%s/%s.pdf' % (figdir, base), bbox_inches='tight')

    return fig, plot


# ------------------------------------------------------------------------------
#
def create_analytical_frame (idx, kind, args, limits, step):
    """
    create an artificial data frame, ie. a data frame which does not contain
    data gathered from an experiment, but data representing an analytical
    construct of some 'kind'.

    idx:    data frame column index to fill (a time column is always created)
    kind:   construct to use (only 'rate' is supporte right now)
    args:   construct specific parameters
    limits: time range for which data are to be created
    step:   time steps for which data are to be created
    """

    import pandas as pd

    # --------------------------------------------------------------------------
    def _frange(start, stop, step):
        while start <= stop:
            yield start
            start += step
    # --------------------------------------------------------------------------
            
    if kind == 'rate' :
        t_0  = args.get ('t_0',  0.0)
        rate = args.get ('rate', 1.0)
        data = list()
        for t in _frange(limits[0], limits[1], step):
            data.append ({'time': t+t_0, idx: t*rate})
        return pd.DataFrame (data)
        
    else:
        raise ValueError ("No such frame kind '%s'" % kind)
        

# ------------------------------------------------------------------------------
#
def add_derived(df):
    """
    Add additional (derived) colums to dataframes
    create columns based on two other columns using an operator
    """
    
    import operator

    # TODO: The fields these are derived from are outdated by now!
    df['executor_queue'] = operator.sub(df['ewo_get'],      df['as_to_ewo'])
    df['raw_runtime']    = operator.sub(df['ewa_complete'], df['ewo_launch'])
    df['full_runtime']   = operator.sub(df['uw_push_done'], df['as_to_ewo'])
    df['watch_delay']    = operator.sub(df['ewa_get'],      df['ewo_to_ewa'])
    df['allocation']     = operator.sub(df['as_allocated'], df['a_to_as'])

    # add a flag to indicate if a unit / pilot / ... is cloned
    # --------------------------------------------------------------------------
    def _cloned (row):
        return 'clone' in row['uid'].lower()
    # --------------------------------------------------------------------------
    df['cloned'] = df.apply(lambda row: _cloned (row), axis=1)

    return df


# ------------------------------------------------------------------------------
#
def add_info(df):
    """
    we also derive some specific info from the event/msg columns, based on
    the mapping defined in _info_entries.  That should make it easier to
    analyse the data.
    """

    import numpy as np

    # --------------------------------------------------------------------------
    def _info (row):
        for info, name, event, msg in _info_entries:
            if  (row['name'] and name  in row['name'] ) and \
                (not event   or  event == row['event']) and \
                (not msg     or  msg   == row['msg']  ):
                return info

        ret = ""
        n   = 0  # 
        for pat, pre in info_names.iteritems():
            if pat in row['name']:
                ret += pre
                n   += 1
                break
        for pat, ev in _info_events.iteritems():
            if ret and pat == row['event']:
                ret += ev
                n   += 1
                break
        for pat, s in _info_pending.iteritems():
            if ret and pat in row['state']:
                ret += s
                break
        # Also create separate info entries for Canceled and Failed units
        for pat, f in _info_premature_final.iteritems():
            if ret and pat in row['state']:
                ret += f
                break

        if ret and n >= 2:
            return ret
        else:
            return np.NaN
    # --------------------------------------------------------------------------
    df['info'] = df.apply(lambda row: _info (row), axis=1)

    return df
    

# ------------------------------------------------------------------------------
#
def get_info_df(df):
    """
    This call will convert a raw profiling frame into a transposed frame where
    possible info tags are columns, and each row contains the respective timings
    for uni unit.  uid is df index.
    """

    import pandas as pd

    uids  = list(df['uid'].unique())
    cols  = set()
    dicts = dict()

    for uid in uids:
        uf_n  = df[df['uid'] == uid]
        tmp1   = uf_n[['time', 'info']].dropna()
        tmp1_d = tmp1.set_index('info').to_dict()['time']

        # add state transitions to dict.  We basically select all rows with
        # a 'state_from', and add a dict entry for the 'state' column.
        # Make sure we ignore transitions where 'state' == 'state_from'.
        tmp2   = uf_n[uf_n['state_from'].notnull()]
        tmp3   = tmp2[tmp2['state'] != tmp2['state_from']]
        tmp4   = tmp3[['time', 'state']].dropna()
        tmp2_d = tmp4.set_index('state').to_dict()['time']
        for k,v in tmp2_d.iteritems():
            tmp1_d[k] = v
        # make sure the frame column headers are complete
        for s in _info_states:
            cols.add(s)

        # make sure we got no info double defined on any unit.  Also derive
        # column names
        l = list(uf_n['info'].dropna())
        for i in set(l):
            cols.add(i)
            if l.count(i)>1:
                raise ValueError('doubled info entry %s (uid:%s)' % (i, uid))

        dicts[uid] = tmp1_d


    new_df = pd.DataFrame(columns=cols, index=uids)
    for uid in dicts:
        new_df.loc[uid] = pd.Series(dicts[uid])

    return new_df


# ------------------------------------------------------------------------------
#
def get_state_df(df):
    """
    This call accepts an 'info' df, and reduces it to only contain state
    transitions.
    """

    return df[_info_states]


# ------------------------------------------------------------------------------
#
def add_states(df):
    """
    Add one additional columns: 'state_from'.  It will have a value for all
    columns where a stateful entity entered a new state, and will have the value
    of the previous state.

    We also fill out the state column, to continue to have the value of any
    previous state setting.
    """
    
    import numpy as np

    # --------------------------------------------------------------------------
    _old_states = dict()
    def _state_from (row):
        old = np.NaN
        if  row['uid']   and \
            row['state'] and \
            row['event'] == 'advance': 
            old = _old_states.get(row['uid'], np.NaN)
            _old_states[row['uid']] = row['state']
        return old
    # --------------------------------------------------------------------------
  # df['state_from'], df['state_to'] = zip(*df.apply(lambda row: _state(row), axis=1))
    df['state_from'] = df.apply(lambda row: _state_from(row), axis=1)

    _old_states = dict()
    # --------------------------------------------------------------------------
    def _state (row):
        if  not row['uid']:
            return np.NaN
        if row['state']:
            _old_states[row['uid']] = row['state']
        return _old_states.get(row['uid'], '')
    # --------------------------------------------------------------------------
    df['state'] = df.apply(lambda row: _state(row), axis=1)

    return df


# ------------------------------------------------------------------------------
#
def get_span(df, spec):
    """
    get the timespan from the first of any of the listed events to the last of
    any of the listed events.  Events are 'info' entries.
    """
    
    import numpy as np

    if not isinstance(spec, list):
        spec = [spec]

    first = float(df.tail(1)['time']) # max
    last  = float(df.head(1)['time']) # min

    ok = False
    for s in spec:
        tmp   = df[df['info'] == s]
        if len(tmp):
            ok = True
            first = min(first, float(tmp.head(1).iloc[0]['time']))
            last  = max(last,  float(tmp.tail(1).iloc[0]['time']))

    if not ok:
        return None

    return last - first


# ------------------------------------------------------------------------------




"""
.. module:: radical.pilot.controller.pilot_launcher_worker
.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import time
import saga
import bson
import pprint
import traceback
import thread
import threading

import weakref
from multiprocessing import Pool

import radical.utils as ru

from ..states       import *
from ..utils        import logger
from ..utils        import timestamp
from ..db.database  import COMMAND_CANCEL_PILOT

from .pilot_launcher_worker import PilotLauncherWorker

import saga.utils.pty_shell as sup

IDLE_TIME  = 1.0  # seconds to sleep after idle cycles


# ----------------------------------------------------------------------------
#
class PilotManagerController(threading.Thread):
    """PilotManagerController is a threading worker that handles backend
       interaction for the PilotManager and Pilot classes.
    """

    # ------------------------------------------------------------------------
    #
    def __init__(self, pmgr_uid, pilot_manager_data, 
        session, pilot_launcher_workers=1):
        """Le constructeur.
        """
        self._session = session

        # The MongoDB database handle.
        self._dbs = self._session.get_dbs()

        # Multithreading stuff
        threading.Thread.__init__(self)

        # Stop event can be set to terminate the main loop
        self._terminate = threading.Event()
        self._terminate.clear()

        # Initialized is set, once the run loop has pulled status
        # at least once. Other functions use it as a guard.
        self._initialized = threading.Event()
        self._initialized.clear()

        # Startup results contains a list of asynchronous startup results.
        self.startup_results = list()
        self.startup_results_lock = threading.Lock()

        # The shard_data_manager handles data exchange between the worker
        # process and the API objects. The communication is unidirectional:
        # workers WRITE to _shared_data and API methods READ from _shared_data.
        # The strucuture of _shared_data is as follows:
        #
        #  self._shared_data[pilot_uid] = {
        #      'data':          pilot_json,
        #      'callbacks':     []
        #      'facade_object': None
        #  }
        #  self._shared_worker_data = {
        #      'job_services':  {url: saga.job.Service}    # dict of job services
        #      'job_ids'     :  {pilot_id : (job_id, url)} # dict of pilot job handles 
        #  }
        #
        self._shared_data = dict()
        self._shared_worker_data = {'job_services' : dict(), 
                                    'job_ids'      : dict()}

        # The manager-level callbacks.
        self._manager_callbacks = list()

        # The different command queues hold pending operations
        # that are passed to the worker. Command queues are inspected during
        # runtime in the run() loop and the worker acts upon them accordingly.
        #
        # Try to register the PilotManager with the database.
        self._pm_id = self._dbs.insert_pilot_manager(
            pmgr_uid=pmgr_uid,
            pilot_manager_data=pilot_manager_data,
            pilot_launcher_workers=pilot_launcher_workers
        )
        self._num_pilot_launcher_workers = pilot_launcher_workers

        # The pilot launcher worker(s) are autonomous processes that
        # execute pilot bootstrap / launcher requests concurrently.
        self._pilot_launcher_worker_pool = []
        for worker_number in range(1, self._num_pilot_launcher_workers+1):
            worker = PilotLauncherWorker(
                session=self._session,
                pilot_manager_id=self._pm_id,
                shared_worker_data=self._shared_worker_data,
                number=worker_number
            )
            self._pilot_launcher_worker_pool.append(worker)
            worker.start()

        self._callback_histories = dict()

    # ------------------------------------------------------------------------
    #
    @classmethod
    def uid_exists(cls, db_connection, pilot_manager_uid):
        """Checks wether a pilot unit manager UID exists.
        """
        exists = False

        if pilot_manager_uid in db_connection.list_pilot_manager_uids():
            exists = True

        return exists

    # ------------------------------------------------------------------------
    #
    @property
    def pilot_manager_uid(self):
        """Returns the uid of the associated PilotMangager
        """
        return self._pm_id

    # ------------------------------------------------------------------------
    #
    def list_pilots(self):
        """List all known pilots.
        """
        return self._dbs.list_pilot_uids(self._pm_id)

    # ------------------------------------------------------------------------
    #
    def get_compute_pilot_data(self, pilot_ids=None):
        """Returns the raw data (json dicts) of one or more ComputePilots
           registered with this Worker / PilotManager
        """
        # Wait for the initialized event to assert proper operation.
        self._initialized.wait()

        try:
            if  pilot_ids is None:
                pilot_ids = self._shared_data.keys ()

            return_list_type = True
            if not isinstance(pilot_ids, list):
                return_list_type = False
                pilot_ids = [pilot_ids]

            data = list()
            for pilot_id in pilot_ids:
                data.append(self._shared_data[pilot_id]['data'])

            if  return_list_type :
                return data
            else :
                return data[0]

        except KeyError as e:
            logger.exception ("Unknown Pilot ID %s : %s" % (pilot_id, e))
            raise

    # ------------------------------------------------------------------------
    #
    def disable_launcher(self):
        """disable pilot launching
        """
        for worker in self._pilot_launcher_worker_pool:
            logger.debug("pworker %s disables launcher %s" % (self.name, worker.name))
            worker.disable()
            logger.debug("pworker %s disabled launcher %s" % (self.name, worker.name))


    # ------------------------------------------------------------------------
    #
    def cancel_launcher(self):
        """cancel the launcher threads
        """
        for worker in self._pilot_launcher_worker_pool:
            logger.debug("pworker %s stops   launcher %s" % (self.name, worker.name))
            worker.stop ()
            worker.join ()
            logger.debug("pworker %s stopped launcher %s" % (self.name, worker.name))


    # ------------------------------------------------------------------------
    #
    def stop(self):
        """stop() signals the process to finish up and terminate.
        """
        logger.debug("pworker %s stopping" % (self.name))
        self._terminate.set()
        self.join()
        logger.debug("pworker %s stopped" % (self.name))

      # logger.debug("Worker thread (ID: %s[%s]) for PilotManager %s stopped." %
      #             (self.name, self.ident, self._pm_id))

    # ------------------------------------------------------------------------
    #
    def call_callbacks(self, pilot_id, new_state):
        """Wrapper function to call all all relevant callbacks, on pilot-level
        as well as manager-level.
        """

        # this is the point where, at the earliest, the application could have
        # been notified about pilot state changes.  So we record that event.
        if  not pilot_id in self._callback_histories :
            self._callback_histories[pilot_id] = list()
        self._callback_histories[pilot_id].append (
                {'timestamp' : timestamp(), 
                 'state'     : new_state})

        for cb in self._shared_data[pilot_id]['callbacks']:
            cb_func = cb['cb_func']
            cb_data = cb['cb_data']
            try:
                if self._shared_data[pilot_id]['facade_object'] :
                    if cb_data:
                        cb_func(self._shared_data[pilot_id]['facade_object'](), new_state, cb_data)
                    else:
                        cb_func(self._shared_data[pilot_id]['facade_object'](), new_state)
                else :
                    logger.error("Couldn't call callback (no pilot instance)")
            except Exception as e:
                logger.exception("Couldn't call callback function %s" % e)
                raise
            except SystemExit:
                # the callback requested a sys exit.  We don't want the
                # callbacks to get into the way of the shutdown, so we 
                # unregister them all right here
                logger.exception('sys.exit from callback')
                self.unregister_pilot_callback(pilot_id)
                thread.interrupt_main()

        # If we have any manager-level callbacks registered, we
        # call those as well!
        for [cb, cb_data] in self._manager_callbacks:
            try:
                if  self._shared_data[pilot_id]['facade_object'] :
                    if  cb_data :
                        cb(self._shared_data[pilot_id]['facade_object'](), new_state, cb_data)
                    else :
                        cb(self._shared_data[pilot_id]['facade_object'](), new_state)
                else :
                    logger.error("Couldn't call manager callback (no pilot instance)")
            except Exception as e:
                logger.exception("Couldn't call callback function %s" % e)
                raise
            except SystemExit:
                # the callback requested a sys exit.  We don't want the
                # callbacks to get into the way of the shutdown, so we 
                # unregister them all right here
                logger.exception('sys.exit from callback')
                self.unregister_pilot_callback(pilot_id)
                thread.interrupt_main()

        # if we meet a final state, we record the object's callback history for
        # later evalutation
        if new_state in (DONE, FAILED, CANCELED):
            self._dbs.publish_compute_pilot_callback_history (pilot_id, self._callback_histories[pilot_id])


    # ------------------------------------------------------------------------
    #
    def run(self):
        """run() is called when the process is started via
           PilotManagerController.start().
        """

        # make sure to catch sys.exit (which raises SystemExit)
        try :

            logger.debug("Worker thread (ID: %s[%s]) for PilotManager %s started." %
                        (self.name, self.ident, self._pm_id))

            while not self._terminate.is_set():

                # Check and update pilots. This needs to be optimized at
                # some point, i.e., state pulling should be conditional
                # or triggered by a tailable MongoDB cursor, etc.
                pilot_list = self._dbs.get_pilots(pilot_manager_id=self._pm_id)
                action = False

                for pilot in pilot_list:
                    pilot_id = str(pilot["_id"])

                    new_state = pilot["state"]
                    if pilot_id in self._shared_data:
                        old_state = self._shared_data[pilot_id]["data"]["state"]
                    else:
                        old_state = None
                        self._shared_data[pilot_id] = {
                            'data':          pilot,
                            'callbacks':     [],
                            'facade_object': None
                        }

                    self._shared_data[pilot_id]['data'] = pilot

                    # FIXME: *groan* what a hack...  The Canceling state is by
                    # the nature of it not recorded in the database, but only in
                    # the local cache.  So if we see it as old state, we have to
                    # avoid state transitions into non-final states in the cache
                    # at all cost -- so we catch this here specifically
                    no_cb = False
                    if  old_state == CANCELING :
                        if  new_state not in [DONE, FAILED, CANCELED] :
                            # restore old state, making the cache explicitly
                            # different than the DB recorded state
                            self._shared_data[pilot_id]["data"]["state"] = old_state 

                            # do not trigger a state cb!
                            no_cb = True

                    if new_state != old_state :
                        action = True

                        if not no_cb :
                            # On a state change, we fire zee callbacks.
                            logger.info("ComputePilot '%s' state changed from '%s' to '%s'." \
                                            % (pilot_id, old_state, new_state))

                            # The state of the pilot has changed, We call all
                            # pilot-level callbacks to propagate this.  This also
                            # includes communication to the unit scheduler which
                            # may, or may not, cancel the pilot's units.
                            self.call_callbacks(pilot_id, new_state)

                    # If the state is 'DONE', 'FAILED' or 'CANCELED', we also
                    # set the state of the compute unit accordingly (but only
                    # for non-final units)
                    if new_state in [FAILED, DONE, CANCELED]:
                        unit_ids = self._dbs.pilot_list_compute_units(pilot_uid=pilot_id)
                        self._dbs.set_compute_unit_state (
                            unit_ids=unit_ids, 
                            state=CANCELED,
                            src_states=[AGENT_STAGING_INPUT_PENDING,
                                        AGENT_STAGING_INPUT,
                                        ALLOCATING_PENDING,
                                        ALLOCATING,
                                        EXECUTING_PENDING,
                                        EXECUTING,
                                        AGENT_STAGING_OUTPUT_PENDING,
                                        AGENT_STAGING_OUTPUT],
                            log="Pilot '%s' has terminated with state '%s'. CU canceled." % (pilot_id, new_state))

                # After the first iteration, we are officially initialized!
                if not self._initialized.is_set():
                    self._initialized.set()

                # sleep a little if this cycle was idle
                if  not action :
                    time.sleep(IDLE_TIME)

        except SystemExit as e :
            logger.debug("pilot manager controller thread caught system exit -- forcing application shutdown")
            thread.interrupt_main ()

        finally :
            # shut down the autonomous pilot launcher worker(s).  
            # This uses terminate=True, so that on loop errors we'll terminate
            # all pilots.  Note however that this instance had
            # a 'close(terminate=True)' called before, then the stop below is
            # a NOOP, and pilots will continue running.
            for worker in self._pilot_launcher_worker_pool:
                logger.debug("pworker %s stops   launcher %s" % (self.name, worker.name))
                worker.stop ()
                logger.debug("pworker %s stopped launcher %s" % (self.name, worker.name))



    # ------------------------------------------------------------------------
    #
    def register_start_pilot_request(self, pilot, resource_config):
        """Register a new pilot start request with the worker.
        """

        # create a new UID for the pilot
        pilot_uid = ru.generate_id ('pilot')

        # switch endpoint type
        fs_url = saga.Url(resource_config['filesystem_endpoint'])

        # Get the sandbox from either the pilot_desc or resource conf
        if pilot.description.sandbox:
            workdir_raw = pilot.description.sandbox
        else:
            workdir_raw = resource_config.get('default_remote_workdir', "$PWD")

        # If the sandbox contains expandables, we need to resolve those remotely.
        # TODO: Note that this will only work for (gsi)ssh or shell based access mechanisms
        if '$' in workdir_raw or '`' in workdir_raw:
            js_url = saga.Url(resource_config['job_manager_endpoint'])

            # The PTYShell will swallow in the job part of the scheme
            if js_url.scheme.endswith('+ssh'):
                # For remote adaptor usage over shh, use that here
                js_url.scheme = 'ssh'
            elif js_url.scheme.endswith('+gsissh'):
                # For remote adaptor usage over gsissh, use that here
                js_url.scheme = 'gsissh'
            elif js_url.scheme in ['fork', 'ssh', 'gsissh']:
                # Use the scheme as is for non-queuing adaptor mechanisms
                pass
            elif '+' not in js_url.scheme:
                # For local access to queueing systems use fork
                js_url.scheme = 'fork'
            else:
                raise Exception("Are there more flavours we need to support?! (%s)" % js_url.scheme)

            # TODO: Why is this 'translation' required?
            if js_url.port is not None:
                url = "%s://%s:%d/" % (js_url.schema, js_url.host, js_url.port)
            else:
                url = "%s://%s/" % (js_url.schema, js_url.host)

            logger.debug("saga.utils.PTYShell ('%s')" % url)
            shell = sup.PTYShell(url, self._session)

            ret, out, err = shell.run_sync(' echo "WORKDIR: %s"' % workdir_raw)
            if ret == 0 and 'WORKDIR:' in out :
                workdir_expanded = out.split(":")[1].strip()
                logger.debug("Determined remote working directory for %s: '%s'" % (url, workdir_expanded))
            else :
                error_msg = "Couldn't determine remote working directory."
                logger.error(error_msg)
                raise Exception(error_msg)
        else:
            workdir_expanded = workdir_raw

        # At this point we have determined the remote 'pwd'
        fs_url.path = "%s/radical.pilot.sandbox" % workdir_expanded

        # This is the base URL / 'sandbox' for the pilot!
        agent_dir_url = saga.Url("%s/%s-%s/" % (str(fs_url), self._session.uid, pilot_uid))

        # Create a database entry for the new pilot.
        pilot_uid, pilot_json = self._dbs.insert_pilot(
            pilot_uid=pilot_uid,
            pilot_manager_uid=self._pm_id,
            pilot_description=pilot.description,
            pilot_sandbox=str(agent_dir_url), 
            global_sandbox=str(fs_url.path)
            )

        # Create a shared data store entry
        self._shared_data[pilot_uid] = {
            'data':          pilot_json,
            'callbacks':     [],
            'facade_object': weakref.ref(pilot)
        }

        return pilot_uid

    # ------------------------------------------------------------------------
    #
    def register_pilot_callback(self, pilot, cb_func, cb_data=None):
        """Registers a callback function.
        """
        pilot_uid = pilot.uid
        self._shared_data[pilot_uid]['callbacks'].append({'cb_func' : cb_func, 
                                                          'cb_data' : cb_data})

        # Add the facade object if missing, e.g., after a re-connect.
        if  self._shared_data[pilot_uid]['facade_object'] is None:
            self._shared_data[pilot_uid]['facade_object'] = weakref.ref(pilot)

        # Callbacks can only be registered when the ComputeAlready has a
        # state. To partially address this shortcoming we call the callback
        # with the current ComputePilot state as soon as it is registered.
        self.call_callbacks(
            pilot.uid,
            self._shared_data[pilot_uid]["data"]["state"]
        )

    # ------------------------------------------------------------------------
    #
    def unregister_pilot_callback(self, pid, cb_func=None):
        """
        Un-registers a callback function -- or all if cb_func is None
        """
        if not pid in self._shared_data:
            raise ValueError('unknown pilot %s' % pid)

        if cb_func:
            # iterate over copy
            for cb in self._shared_data[pid]['callbacks'][:]: 
                if cb_func == cb['cb_func']:
                    self._shared_data[pid]['callbacks'].remove(cb)
        else:
            # remove all callbacks
            self._shared_data[pid]['callbacks'] = []


    # ------------------------------------------------------------------------
    #
    def register_manager_callback(self, cb_func, cb_data=None):
        """Registers a manager-level callback.
        """
        self._manager_callbacks.append([cb_func, cb_data])

    # ------------------------------------------------------------------------
    #
    def register_cancel_pilots_request(self, pilot_ids=None):
        """Registers one or more pilots for cancelation.
        """

        if pilot_ids is None:

            pilot_ids = list()

            for pilot in self._dbs.get_pilots(pilot_manager_id=self._pm_id) :
                pilot_ids.append (str(pilot["_id"]))


        self._dbs.send_command_to_pilot(COMMAND_CANCEL_PILOT, pilot_ids=pilot_ids)
        logger.info("Sent 'COMMAND_CANCEL_PILOT' command to pilots %s.", pilot_ids)

        # pilots which are in ACTIVE state should now have time to react on the
        # CANCEL command sent above.  Meanwhile, we'll cancel all pending
        # pilots.  If that is done, we wait a little, say 10 seconds, to give
        # the pilot time to pick up the request and shut down -- but if it does
        # not do that, it will get killed the hard way...
        delayed_cancel = list()

        for pilot_id in pilot_ids:
            if  pilot_id in self._shared_data:

                # we don't want to see any callbacks during shutdown
                # FXIME: this is semantically incorrect, as a CANCELED state cb
                #        should still be issued.  But, shutdown races screw us
                #        once more...
                self.unregister_pilot_callback(pilot_id)

                # read state from _shared_data only once, so that it does not
                # change under us...
                old_state = str(self._shared_data[pilot_id]["data"]["state"])

                if old_state in [DONE, FAILED, CANCELED]:
                    logger.debug("can't actively cancel pilot %s: already in final state" % pilot_id)

                elif old_state in [PENDING_LAUNCH, LAUNCHING, PENDING_ACTIVE]:

                    if pilot_id in self._shared_worker_data['job_ids']:

                        try:
                            job_id, js_url = self._shared_worker_data['job_ids'][pilot_id]
                            self._shared_data[pilot_id]["data"]["state"] = CANCELING
                            logger.info("actively cancel pilot %s (%s, %s)" % (pilot_id, job_id, js_url))

                            js = self._shared_worker_data['job_services'][js_url]
                            job = js.get_job(job_id)
                            job.cancel()
                        except Exception as e:
                            logger.exception('pilot cancelation failed')


                    else:
                        logger.debug("can't (yet) cancel starting pilot %s" % pilot_id)
                        delayed_cancel.append(pilot_id)

                else:
                    logger.debug("delay to actively cancel pilot %s: state %s" % (pilot_id, old_state))
                    delayed_cancel.append(pilot_id)

            else:
                raise RuntimeError("unknown pilot" % pilot_id)

        # now tend to all delayed cancellation requests (ie. active pilots) --
        # if there are any
        if  delayed_cancel :

            # grant some levay (30 sec) to the unruly children...
            delay = 30
            start = time.time()
            logger.report.idle(mode='start')
            logger.report.idle(c=' ')

            # we idle as long as any pilot is left to watch and we did not yet
            # hit the delay timeout
            while delayed_cancel and time.time() - start < delay:

                logger.report.idle()
                for pilot_id in delayed_cancel[:]:

                    if pilot_id not in self._shared_data:
                        continue

                    state = self._shared_data[pilot_id]["data"]["state"]
                    if state in [DONE, FAILED, CANCELED]:
                        delayed_cancel.remove(pilot_id)

                time.sleep(0.3)
            logger.report.idle(mode='stop')

            # force-kill any remaining pilots
            for pilot_id in delayed_cancel :

                if pilot_id in self._shared_worker_data['job_ids'] :

                    try :
                        # FIXME: the SAGA layer may raise an exception, because
                        # there is a race between the delay state check above
                        # and the cancel below, and the pilot may be gone.  That
                        # is not a problem, we can ignore it -- what is
                        # a problem is that the SAGA logger will complain rather
                        # loundly...
                        job_id, js_url = self._shared_worker_data['job_ids'][pilot_id]
                        logger.info ("actively cancel pilot %s (delayed) (%s, %s)" % (pilot_id, job_id, js_url))

                        js = self._shared_worker_data['job_services'][js_url]
                        job = js.get_job (job_id)
                        job.cancel ()
                    except Exception as e :
                        logger.info('delayed pilot cancelation failed. '
                                    'This is not necessarily a problem.')

                else :
                    logger.error("can't actively cancel pilot %s: no job id known (delayed)" % pilot_id)
                    logger.debug (pprint.pformat (self._shared_worker_data))



# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


from .base import LaunchMethod


# ==============================================================================
#
class MPIRunDPlace(LaunchMethod):
    # TODO: This needs both mpirun and dplace

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, logger):

        LaunchMethod.__init__(self, cfg, logger)


    # --------------------------------------------------------------------------
    #
    def _configure(self):
        # dplace: job launcher for SGI systems (e.g. on Blacklight)
        self.launch_command = self._which('dplace')
        self.mpirun_command = self._which('mpirun')


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        opaque_slots = cu['opaque_slots']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_cores   = cud['cores']
        task_args    = cud.get('arguments') or []
        task_argstr  = self._create_arg_string(task_args)

        if not 'task_offsets' in opaque_slots:
            raise RuntimeError('insufficient information to launch via %s: %s' \
                    % (self.name, opaque_slots))

        task_offsets = opaque_slots['task_offsets']

        if task_argstr:
            task_command = "%s %s" % (task_exec, task_argstr)
        else:
            task_command = task_exec

        dplace_offset = task_offsets

        mpirun_dplace_command = "%s -np %d %s -c %d-%d %s" % \
            (self.mpirun_command, task_cores, self.launch_command,
             dplace_offset, dplace_offset+task_cores-1, task_command)

        return mpirun_dplace_command, None


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import radical.utils as ru

from ... import utils     as rpu
from ... import states    as rps
from ... import constants as rpc


# ------------------------------------------------------------------------------
# 'enum' for RP's agent staging input types
RP_ASI_NAME_DEFAULT = "DEFAULT"


# ==============================================================================
#
class AgentStagingInputComponent(rpu.Component):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg):

        rpu.Component.__init__(self, rpc.AGENT_STAGING_INPUT_COMPONENT, cfg)


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Spawner
    #
    @classmethod
    def create(cls, cfg):

        name = cfg.get('agent_stagin_input_component', RP_ASI_NAME_DEFAULT)

        # Make sure that we are the base-class!
        if cls != AgentStagingInputComponent:
            raise TypeError("Factory only available to base class!")

        from .default import Default

        try:
            impl = {
                RP_ASI_NAME_DEFAULT: Default
            }[name]

            impl = impl(cfg)
            return impl

        except KeyError:
            raise ValueError("AgentStagingInputComponent '%s' unknown or defunct" % name)


# ------------------------------------------------------------------------------





# ------------------------------------------------------------------------------
#
# global constants
#
UMGR_SCHEDULING_QUEUE          = 'umgr_scheduling_queue'
UMGR_STAGING_INPUT_QUEUE       = 'umgr_staging_input_queue'
UMGR_STAGING_OUTPUT_QUEUE      = 'umgr_staging_output_queue'

UMGR_SCHEDULING_COMPONENT      = 'UmgrSchedulingComponent'
UMGR_STAGING_INPUT_COMPONENT   = 'UmgrStagingInputComponent'
UMGR_STAGING_OUTPUT_COMPONENT  = 'UmgrStagingOutputComponent'
UMGR_UPDATE_WORKER             = 'UmgrUpdateWorker'

AGENT_STAGING_INPUT_QUEUE      = 'agent_staging_input_queue'
AGENT_SCHEDULING_QUEUE         = 'agent_scheduling_queue'
AGENT_EXECUTING_QUEUE          = 'agent_executing_queue'
AGENT_STAGING_OUTPUT_QUEUE     = 'agent_staging_output_queue'

AGENT_STAGING_INPUT_COMPONENT  = 'AgentStagingInputComponent'
AGENT_SCHEDULING_COMPONENT     = 'AgentSchedulingComponent'
AGENT_EXECUTING_COMPONENT      = 'AgentExecutingComponent'
AGENT_STAGING_OUTPUT_COMPONENT = 'AgentStagingOutputComponent'
AGENT_UPDATE_WORKER            = 'AgentUpdateWorker'
AGENT_HEARTBEAT_WORKER         = 'AgentHeartbeatWorker'

PMGR_LAUNCHING_QUEUE           = 'pmgr_launching_queue'

PMGR_LAUNCHING_COMPONENT       = 'PmgrLaunchingComponent'

UMGR_STATE_PUBSUB              = 'umgr_state_pubsub'

AGENT_UNSCHEDULE_PUBSUB        = 'agent_unschedule_pubsub'
AGENT_RESCHEDULE_PUBSUB        = 'agent_reschedule_pubsub'
AGENT_COMMAND_PUBSUB           = 'agent_command_pubsub'
AGENT_STATE_PUBSUB             = 'agent_state_pubsub'


# ------------------------------------------------------------------------------
#
# protocol defines
#
COMMAND_CANCEL_PILOT        = "Cancel_Pilot"
COMMAND_CANCEL_COMPUTE_UNIT = "Cancel_Compute_Unit"
COMMAND_KEEP_ALIVE          = "Keep_Alive"
COMMAND_FIELD               = "commands"
COMMAND_TYPE                = "type"
COMMAND_ARG                 = "arg"
COMMAND_CANCEL              = "Cancel"
COMMAND_SCHEDULE            = "schedule"
COMMAND_RESCHEDULE          = "reschedule"
COMMAND_UNSCHEDULE          = "unschedule"
COMMAND_WAKEUP              = "wakeup"

# two-state for resource occupation.
FREE = 'Free'
BUSY = 'Busy'


# ------------------------------------------------------------------------------
#
# staging defines
#
COPY     = 'Copy'     # local cp
LINK     = 'Link'     # local ln -s
MOVE     = 'Move'     # local mv
TRANSFER = 'Transfer' # saga remote transfer TODO: This might just be a special case of copy

#
# Flags
#
CREATE_PARENTS = 'CreateParents'  # Create parent directories while writing file
SKIP_FAILED    = 'SkipFailed'     # Don't stage out files if tasks failed

#
# Defaults
#
DEFAULT_ACTION   = TRANSFER
DEFAULT_PRIORITY = 0
DEFAULT_FLAGS    = [CREATE_PARENTS, SKIP_FAILED]
STAGING_AREA     = 'staging_area'





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import copy
import stat
import time
import Queue
import tempfile
import threading
import traceback
import subprocess

import radical.utils as ru

from .... import pilot     as rp
from ...  import utils     as rpu
from ...  import states    as rps
from ...  import constants as rpc

from .base import AgentExecutingComponent


# ==============================================================================
#
class Popen(AgentExecutingComponent) :

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg):

        AgentExecutingComponent.__init__ (self, cfg)


    # --------------------------------------------------------------------------
    #
    def initialize_child(self):

      # self.declare_input (rps.AGENT_EXECUTING_PENDING, rpc.AGENT_EXECUTING_QUEUE)
      # self.declare_worker(rps.AGENT_EXECUTING_PENDING, self.work)

        self.declare_input (rps.EXECUTING_PENDING, rpc.AGENT_EXECUTING_QUEUE)
        self.declare_worker(rps.EXECUTING_PENDING, self.work)

        self.declare_output(rps.AGENT_STAGING_OUTPUT_PENDING, rpc.AGENT_STAGING_OUTPUT_QUEUE)

        self.declare_publisher ('unschedule', rpc.AGENT_UNSCHEDULE_PUBSUB)
        self.declare_publisher ('state',      rpc.AGENT_STATE_PUBSUB)

        # all components use the command channel for control messages
        self.declare_publisher ('command', rpc.AGENT_COMMAND_PUBSUB)
        self.declare_subscriber('command', rpc.AGENT_COMMAND_PUBSUB, self.command_cb)

        self._cancel_lock    = threading.RLock()
        self._cus_to_cancel  = list()
        self._cus_to_watch   = list()
        self._watch_queue    = Queue.Queue ()

        self._pilot_id = self._cfg['pilot_id']

        # run watcher thread
        self._terminate = threading.Event()
        self._watcher   = threading.Thread(target=self._watch, name="Watcher")
        self._watcher.daemon = True
        self._watcher.start ()

        # The AgentExecutingComponent needs the LaunchMethods to construct
        # commands.
        self._task_launcher = rp.agent.LM.create(
                name   = self._cfg.get('task_launch_method'),
                cfg    = self._cfg,
                logger = self._log)

        self._mpi_launcher = rp.agent.LM.create(
                name   = self._cfg.get('mpi_launch_method'),
                cfg    = self._cfg,
                logger = self._log)

        # communicate successful startup
        self.publish('command', {'cmd' : 'alive',
                                 'arg' : self.cname})

        self._cu_environment = self._populate_cu_environment()

        self.tmpdir = tempfile.gettempdir()


    # --------------------------------------------------------------------------
    #
    def finalize_child(self):

        # terminate watcher thread
        self._terminate.set()
        self._watcher.join()

        # communicate finalization
        self.publish('command', {'cmd' : 'final',
                                 'arg' : self.cname})


    # --------------------------------------------------------------------------
    #
    def command_cb(self, topic, msg):

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd == 'cancel_unit':

            self._log.info("cancel unit command (%s)" % arg)
            with self._cancel_lock:
                self._cus_to_cancel.append(arg)


    # --------------------------------------------------------------------------
    #
    def _populate_cu_environment(self):
        """Derive the environment for the cu's from our own environment."""

        # Get the environment of the agent
        new_env = copy.deepcopy(os.environ)

        #
        # Mimic what virtualenv's "deactivate" would do
        #
        old_path = new_env.pop('_OLD_VIRTUAL_PATH', None)
        if old_path:
            new_env['PATH'] = old_path

        old_home = new_env.pop('_OLD_VIRTUAL_PYTHONHOME', None)
        if old_home:
            new_env['PYTHON_HOME'] = old_home

        old_ps = new_env.pop('_OLD_VIRTUAL_PS1', None)
        if old_ps:
            new_env['PS1'] = old_ps

        new_env.pop('VIRTUAL_ENV', None)

        # Remove the configured set of environment variables from the
        # environment that we pass to Popen.
        for e in new_env.keys():
            env_removables = list()
            if self._mpi_launcher : env_removables += self._mpi_launcher.env_removables
            if self._task_launcher: env_removables += self._task_launcher.env_removables
            for r in  env_removables:
                if e.startswith(r):
                    new_env.pop(e, None)

        return new_env


    # --------------------------------------------------------------------------
    #
    def work(self, cu):

      # self.advance(cu, rps.AGENT_EXECUTING, publish=True, push=False)
        self.advance(cu, rps.EXECUTING, publish=True, push=False)

        try:
            if cu['description']['mpi']:
                launcher = self._mpi_launcher
            else :
                launcher = self._task_launcher

            if not launcher:
                raise RuntimeError("no launcher (mpi=%s)" % cu['description']['mpi'])

            self._log.debug("Launching unit with %s (%s).", launcher.name, launcher.launch_command)

            assert(cu['opaque_slots']) # FIXME: no assert, but check
            self._prof.prof('exec', msg='unit launch', uid=cu['_id'])

            # Start a new subprocess to launch the unit
            self.spawn(launcher=launcher, cu=cu)

        except Exception as e:
            # append the startup error to the units stderr.  This is
            # not completely correct (as this text is not produced
            # by the unit), but it seems the most intuitive way to
            # communicate that error to the application/user.
            self._log.exception("error running CU")
            cu['stderr'] += "\nPilot cannot start compute unit:\n%s\n%s" \
                            % (str(e), traceback.format_exc())

            # Free the Slots, Flee the Flots, Ree the Frots!
            if cu['opaque_slots']:
                self.publish('unschedule', cu)

            self.advance(cu, rps.FAILED, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def spawn(self, launcher, cu):

        self._prof.prof('spawn', msg='unit spawn', uid=cu['_id'])

        if False:
            cu_tmpdir = '%s/%s' % (self.tmpdir, cu['_id'])
        else:
            cu_tmpdir = cu['workdir']

        rpu.rec_makedir(cu_tmpdir)
        launch_script_name = '%s/radical_pilot_cu_launch_script.sh' % cu_tmpdir
        self._log.debug("Created launch_script: %s", launch_script_name)

        with open(launch_script_name, "w") as launch_script:
            launch_script.write('#!/bin/sh\n\n')

            if 'RADICAL_PILOT_PROFILE' in os.environ:
                launch_script.write("echo script start_script `%s` >> %s/PROF\n" % (cu['gtod'], cu_tmpdir))
            launch_script.write('\n# Change to working directory for unit\ncd %s\n' % cu_tmpdir)
            if 'RADICAL_PILOT_PROFILE' in os.environ:
                launch_script.write("echo script after_cd `%s` >> %s/PROF\n" % (cu['gtod'], cu_tmpdir))

            # Before the Big Bang there was nothing
            if cu['description']['pre_exec']:
                pre_exec_string = ''
                if isinstance(cu['description']['pre_exec'], list):
                    for elem in cu['description']['pre_exec']:
                        pre_exec_string += "%s\n" % elem
                else:
                    pre_exec_string += "%s\n" % cu['description']['pre_exec']
                # Note: extra spaces below are for visual alignment
                launch_script.write("# Pre-exec commands\n")
                if 'RADICAL_PILOT_PROFILE' in os.environ:
                    launch_script.write("echo pre  start `%s` >> %s/PROF\n" % (cu['gtod'], cu_tmpdir))
                launch_script.write(pre_exec_string)
                if 'RADICAL_PILOT_PROFILE' in os.environ:
                    launch_script.write("echo pre  stop `%s` >> %s/PROF\n" % (cu['gtod'], cu_tmpdir))

            # Create string for environment variable setting
            env_string = 'export'
            if cu['description']['environment']:
                for key,val in cu['description']['environment'].iteritems():
                    env_string += ' %s=%s' % (key, val)
            env_string += " RP_SESSION_ID=%s" % self._cfg['session_id']
            env_string += " RP_PILOT_ID=%s"   % self._cfg['pilot_id']
            env_string += " RP_AGENT_ID=%s"   % self._cfg['agent_name']
            env_string += " RP_SPAWNER_ID=%s" % self.cname
            env_string += " RP_UNIT_ID=%s"    % cu['_id']
            launch_script.write('# Environment variables\n%s\n' % env_string)

            # The actual command line, constructed per launch-method
            try:
                launch_command, hop_cmd = launcher.construct_command(cu, launch_script_name)

                if hop_cmd : cmdline = hop_cmd
                else       : cmdline = launch_script_name

            except Exception as e:
                msg = "Error in spawner (%s)" % e
                self._log.exception(msg)
                raise RuntimeError(msg)

            launch_script.write("# The command to run\n")
            launch_script.write("%s\n" % launch_command)
            launch_script.write("RETVAL=$?\n")
            if 'RADICAL_PILOT_PROFILE' in os.environ:
                launch_script.write("echo script after_exec `%s` >> %s/PROF\n" % (cu['gtod'], cu_tmpdir))

            # After the universe dies the infrared death, there will be nothing
            if cu['description']['post_exec']:
                post_exec_string = ''
                if isinstance(cu['description']['post_exec'], list):
                    for elem in cu['description']['post_exec']:
                        post_exec_string += "%s\n" % elem
                else:
                    post_exec_string += "%s\n" % cu['description']['post_exec']
                launch_script.write("# Post-exec commands\n")
                if 'RADICAL_PILOT_PROFILE' in os.environ:
                    launch_script.write("echo post start `%s` >> %s/PROF\n" % (cu['gtod'], cu_tmpdir))
                launch_script.write('%s\n' % post_exec_string)
                if 'RADICAL_PILOT_PROFILE' in os.environ:
                    launch_script.write("echo post stop  `%s` >> %s/PROF\n" % (cu['gtod'], cu_tmpdir))

            launch_script.write("# Exit the script with the return code from the command\n")
            launch_script.write("exit $RETVAL\n")

        # done writing to launch script, get it ready for execution.
        st = os.stat(launch_script_name)
        os.chmod(launch_script_name, st.st_mode | stat.S_IEXEC)
        self._prof.prof('command', msg='launch script constructed', uid=cu['_id'])

        _stdout_file_h = open(cu['stdout_file'], "w")
        _stderr_file_h = open(cu['stderr_file'], "w")
        self._prof.prof('command', msg='stdout and stderr files created', uid=cu['_id'])

        self._log.info("Launching unit %s via %s in %s", cu['_id'], cmdline, cu_tmpdir)

        proc = subprocess.Popen(args               = cmdline,
                                bufsize            = 0,
                                executable         = None,
                                stdin              = None,
                                stdout             = _stdout_file_h,
                                stderr             = _stderr_file_h,
                                preexec_fn         = None,
                                close_fds          = True,
                                shell              = True,
                                cwd                = cu_tmpdir,
                                env                = self._cu_environment,
                                universal_newlines = False,
                                startupinfo        = None,
                                creationflags      = 0)

        self._prof.prof('spawn', msg='spawning passed to popen', uid=cu['_id'])

        cu['started'] = rpu.timestamp()
        cu['proc']    = proc

        self._watch_queue.put(cu)


    # --------------------------------------------------------------------------
    #
    def _watch(self):

        self._prof.prof('run', uid=self._pilot_id)
        try:

            while not self._terminate.is_set():

                cus = list()

                try:

                    # we don't want to only wait for one CU -- then we would
                    # pull CU state too frequently.  OTOH, we also don't want to
                    # learn about CUs until all slots are filled, because then
                    # we may not be able to catch finishing CUs in time -- so
                    # there is a fine balance here.  Balance means 100 (FIXME).
                  # self._prof.prof('ExecWorker popen watcher pull cu from queue')
                    MAX_QUEUE_BULKSIZE = 100
                    while len(cus) < MAX_QUEUE_BULKSIZE :
                        cus.append (self._watch_queue.get_nowait())

                except Queue.Empty:

                    # nothing found -- no problem, see if any CUs finished
                    pass

                # add all cus we found to the watchlist
                for cu in cus :

                    self._prof.prof('passed', msg="ExecWatcher picked up unit", uid=cu['_id'])
                    self._cus_to_watch.append (cu)

                # check on the known cus.
                action = self._check_running()

                if not action and not cus :
                    # nothing happened at all!  Zzz for a bit.
                    time.sleep(self._cfg['db_poll_sleeptime'])

        except Exception as e:
            self._log.exception("Error in ExecWorker watch loop (%s)" % e)
            # FIXME: this should signal the ExecWorker for shutdown...


    # --------------------------------------------------------------------------
    # Iterate over all running tasks, check their status, and decide on the
    # next step.  Also check for a requested cancellation for the tasks.
    def _check_running(self):

        action = 0

        for cu in self._cus_to_watch:

            # poll subprocess object
            exit_code = cu['proc'].poll()
            now       = rpu.timestamp()

            if exit_code is None:
                # Process is still running

                if cu['_id'] in self._cus_to_cancel:

                    # FIXME: there is a race condition between the state poll
                    # above and the kill command below.  We probably should pull
                    # state after kill again?

                    # We got a request to cancel this cu
                    action += 1
                    cu['proc'].kill()
                    cu['proc'].wait() # make sure proc is collected

                    with self._cancel_lock:
                        self._cus_to_cancel.remove(cu['_id'])

                    self._prof.prof('final', msg="execution canceled", uid=cu['_id'])

                    del(cu['proc'])  # proc is not json serializable
                    self.publish('unschedule', cu)
                    self.advance(cu, rps.CANCELED, publish=True, push=False)

                    # we don't need to watch canceled CUs
                    self._cus_to_watch.remove(cu)

            else:
                self._prof.prof('exec', msg='execution complete', uid=cu['_id'])

                # make sure proc is collected
                cu['proc'].wait()

                # we have a valid return code -- unit is final
                action += 1
                self._log.info("Unit %s has return code %s.", cu['_id'], exit_code)

                cu['exit_code'] = exit_code
                cu['finished']  = now

                # Free the Slots, Flee the Flots, Ree the Frots!
                self._cus_to_watch.remove(cu)
                del(cu['proc'])  # proc is not json serializable
                self.publish('unschedule', cu)

                if exit_code != 0:
                    # The unit failed - fail after staging output
                    self._prof.prof('final', msg="execution failed", uid=cu['_id'])
                    cu['target_state'] = rps.FAILED

                else:
                    # The unit finished cleanly, see if we need to deal with
                    # output data.  We always move to stageout, even if there are no
                    # directives -- at the very least, we'll upload stdout/stderr
                    self._prof.prof('final', msg="execution succeeded", uid=cu['_id'])
                    cu['target_state'] = rps.DONE

                self.advance(cu, rps.AGENT_STAGING_OUTPUT_PENDING, publish=True, push=True)

        return action


